# Copyright (c) 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Defines database models for Barbican
"""
import hashlib

from oslo_serialization import jsonutils as json
from oslo_utils import timeutils
import sqlalchemy as sa
from sqlalchemy.ext import compiler
from sqlalchemy.ext import declarative
from sqlalchemy import orm
from sqlalchemy.orm import collections as col
from sqlalchemy import types as sql_types

from barbican.common import exception
from barbican.common import utils
from barbican import i18n as u

BASE = declarative.declarative_base()
ERROR_REASON_LENGTH = 255
SUB_STATUS_LENGTH = 36
SUB_STATUS_MESSAGE_LENGTH = 255


# Allowed entity states
class States(object):
    PENDING = 'PENDING'
    ACTIVE = 'ACTIVE'
    ERROR = 'ERROR'

    @classmethod
    def is_valid(cls, state_to_test):
        """Tests if a state is a valid one."""
        return state_to_test in cls.__dict__


class OrderType(object):
    KEY = 'key'
    ASYMMETRIC = 'asymmetric'

    @classmethod
    def is_valid(cls, order_type):
        """Tests if a order type is a valid one."""
        return order_type in cls.__dict__


class OrderStatus(object):
    def __init__(self, id, message):
        self.id = id
        self.message = message


@compiler.compiles(sa.BigInteger, 'sqlite')
def compile_big_int_sqlite(type_, compiler, **kw):
    return 'INTEGER'


class JsonBlob(sql_types.TypeDecorator):
    """JsonBlob is custom type for fields which need to store JSON text."""
    impl = sa.Text

    def process_bind_param(self, value, dialect):
        if value is not None:
            return json.dumps(value)
        return value

    def process_result_value(self, value, dialect):
        if value is not None:
            return json.loads(value)
        return value


class ModelBase(object):
    """Base class for Nova and Barbican Models."""
    __table_args__ = {'mysql_engine': 'InnoDB'}
    __table_initialized__ = False
    __protected_attributes__ = {
        "created_at", "updated_at", "deleted_at", "deleted",
    }

    id = sa.Column(
        sa.String(36),
        primary_key=True,
        default=utils.generate_uuid)
    created_at = sa.Column(
        sa.DateTime,
        default=timeutils.utcnow,
        nullable=False)
    updated_at = sa.Column(
        sa.DateTime,
        default=timeutils.utcnow,
        nullable=False,
        onupdate=timeutils.utcnow)
    deleted_at = sa.Column(sa.DateTime)
    deleted = sa.Column(sa.Boolean, nullable=False, default=False)
    status = sa.Column(sa.String(20), nullable=False, default=States.PENDING)

    def save(self, session=None):
        """Save this object."""
        # import api here to prevent circular dependency problem
        import barbican.model.repositories
        session = session or barbican.model.repositories.get_session()
        # if model is being created ensure that created/updated are the same
        if self.id is None:
            self.created_at = timeutils.utcnow()
            self.updated_at = self.created_at
        session.add(self)
        session.flush()

    def delete(self, session=None):
        """Delete this object."""
        import barbican.model.repositories
        session = session or barbican.model.repositories.get_session()
        self._do_delete_children(session)
        session.delete(self)

    def get(self, key, default=None):
        return getattr(self, key, default)

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        pass

    def update(self, values):
        """dict.update() behaviour."""
        for k, v in values.items():
            self[k] = v

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __getitem__(self, key):
        return getattr(self, key)

    def __iter__(self):
        self._i = iter(orm.object_mapper(self).sa.Columns)
        return self

    def next(self):
        n = next(self._i).name
        return n, getattr(self, n)

    def keys(self):
        return self.__dict__.keys()

    def values(self):
        return self.__dict__.values()

    def items(self):
        return self.__dict__.items()

    def to_dict(self):
        return self.__dict__.copy()

    def to_dict_fields(self):
        """Returns a dictionary of just the db fields of this entity."""

        if self.created_at:
            created_at = self.created_at.isoformat()
        else:
            created_at = self.created_at

        if self.updated_at:
            updated_at = self.updated_at.isoformat()
        else:
            updated_at = self.updated_at

        dict_fields = {
            'created': created_at,
            'updated': updated_at,
            'status': self.status
        }

        if self.deleted_at:
            dict_fields['deleted_at'] = self.deleted_at.isoformat()
        if self.deleted:
            dict_fields['deleted'] = True
        dict_fields.update(self._do_extra_dict_fields())
        return dict_fields

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {}

    def _iso_to_datetime(self, expiration):
        """Convert ISO formatted string to datetime."""
        if isinstance(expiration, str):
            expiration_iso = timeutils.parse_isotime(expiration.strip())
            expiration = timeutils.normalize_time(expiration_iso)

        return expiration


class SoftDeleteMixIn(object):
    """Mix-in class that adds soft delete functionality."""

    def delete(self, session=None):
        """Delete this object."""
        import barbican.model.repositories
        session = session or barbican.model.repositories.get_session()
        self.deleted = True
        self.deleted_at = timeutils.utcnow()
        self.save(session=session)

        self._do_delete_children(session)


class ContainerSecret(BASE, SoftDeleteMixIn, ModelBase):
    """Represents an association between a Container and a Secret."""

    __tablename__ = 'container_secret'

    name = sa.Column(sa.String(255), nullable=True)
    container_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('containers.id'),
        index=True,
        nullable=False)
    secret_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secrets.id'),
        index=True,
        nullable=False)

    # Eager load this relationship via 'lazy=False'.
    container = orm.relationship(
        'Container',
        back_populates='container_secrets',
        primaryjoin='and_(ContainerSecret.container_id == Container.id, ContainerSecret.deleted != True)',  # noqa: E501
        lazy=False,
    )
    secrets = orm.relationship(
        'Secret',
        back_populates='container_secrets',
        primaryjoin='and_(ContainerSecret.secret_id == Secret.id, ContainerSecret.deleted != True)',  # noqa: E501
        lazy=False,
    )

    __table_args__ = (
        sa.UniqueConstraint(
            'container_id', 'secret_id', 'name',
            name='_container_secret_name_uc'),
    )

    def __init__(self, check_exc=True):
        super(ContainerSecret, self).__init__()


class Project(BASE, SoftDeleteMixIn, ModelBase):
    """Represents a Project in the datastore.

    Projects are users that wish to store secret information within Barbican.
    """

    __tablename__ = 'projects'

    external_id = sa.Column(sa.String(255), unique=True)

    orders = orm.relationship('Order', back_populates='project')
    secrets = orm.relationship('Secret', back_populates='project')
    keks = orm.relationship('KEKDatum', back_populates='project')
    containers = orm.relationship('Container', back_populates='project')
    cas = orm.relationship(
        'ProjectCertificateAuthority',
        back_populates='project')
    project_quotas = orm.relationship(
        'ProjectQuotas',
        back_populates='project')
    preferred_ca = orm.relationship(
        'PreferredCertificateAuthority',
        back_populates='project')
    preferred_secret_store = orm.relationship(
        'ProjectSecretStore',
        back_populates='project')

    def __init__(self, check_exc=True):
        super(Project, self).__init__()

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'external_id': self.external_id}


class Secret(BASE, SoftDeleteMixIn, ModelBase):
    """Represents a Secret in the datastore.

    Secrets are any information Projects wish to store within
    Barbican, though the actual encrypted data is stored in one
    or more EncryptedData entities on behalf of a Secret.
    """

    __tablename__ = 'secrets'

    name = sa.Column(sa.String(255))
    secret_type = sa.Column(
        sa.String(255),
        server_default=utils.SECRET_TYPE_OPAQUE)
    expiration = sa.Column(sa.DateTime, default=None)
    algorithm = sa.Column(sa.String(255))
    bit_length = sa.Column(sa.Integer)
    mode = sa.Column(sa.String(255))
    creator_id = sa.Column(sa.String(255))
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id', name='secrets_project_fk'),
        index=True,
        nullable=False)

    # TODO(jwood): Performance - Consider avoiding full load of all
    #   datum attributes here. This is only being done to support the
    #   building of the list of supported content types when secret
    #   metadata is retrieved.
    #   See barbican.api.resources.py::SecretsResource.on_get()
    # Eager load this relationship via 'lazy=False'.
    encrypted_data = orm.relationship("EncryptedDatum", lazy=False)
    project = orm.relationship('Project', back_populates='secrets')
    container_secrets = orm.relationship(
        "ContainerSecret",
        primaryjoin='and_(ContainerSecret.secret_id==Secret.id, ContainerSecret.deleted!=True)',  # noqa: E501
        back_populates='secrets',
        lazy=False)
    secret_store_metadata = orm.relationship(
        "SecretStoreMetadatum",
        collection_class=col.attribute_mapped_collection('key'),
        back_populates="secret",
        cascade="all, delete-orphan",
        cascade_backrefs=False)
    secret_user_metadata = orm.relationship(
        "SecretUserMetadatum",
        collection_class=col.attribute_mapped_collection('key'),
        back_populates="secret",
        cascade="all, delete-orphan",
        cascade_backrefs=False)
    consumers = orm.relationship(
        "SecretConsumerMetadatum",
        back_populates="secret",
        cascade="all, delete-orphan",
        cascade_backrefs=False)
    secret_acls = orm.relationship(
        "SecretACL",
        back_populates="secret",
        lazy=False)

    def __init__(self, parsed_request=None, check_exc=True):
        """Creates secret from a dict."""
        super(Secret, self).__init__()

        if parsed_request:
            self.name = parsed_request.get('name')
            self.secret_type = parsed_request.get(
                'secret_type',
                utils.SECRET_TYPE_OPAQUE)
            expiration = self._iso_to_datetime(parsed_request.get
                                               ('expiration'))
            self.expiration = expiration
            self.algorithm = parsed_request.get('algorithm')
            self.bit_length = parsed_request.get('bit_length')
            self.mode = parsed_request.get('mode')
            self.creator_id = parsed_request.get('creator_id')
            self.project_id = parsed_request.get('project_id')

        self.status = States.ACTIVE

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        for k, v in self.secret_store_metadata.items():
            v.delete(session)

        for k, v in self.secret_user_metadata.items():
            v.delete(session)

        for datum in self.encrypted_data:
            datum.delete(session)

        for secret_ref in self.container_secrets:
            secret_ref.delete(session)

        for secret_acl in self.secret_acls:
            secret_acl.delete(session)

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        if self.expiration:
            expiration = self.expiration.isoformat()
        else:
            expiration = self.expiration

        return {
            'secret_id': self.id,
            'name': self.name,
            'secret_type': self.secret_type,
            'expiration': expiration,
            'algorithm': self.algorithm,
            'bit_length': self.bit_length,
            'mode': self.mode,
            'creator_id': self.creator_id,
            "consumers": [
                {
                    "service": consumer.service,
                    "resource_type": consumer.resource_type,
                    "resource_id": consumer.resource_id,
                } for consumer in self.consumers if not consumer.deleted
            ],
        }


class SecretStoreMetadatum(BASE, SoftDeleteMixIn, ModelBase):
    """Represents Secret Store metadatum for a single key-value pair."""

    __tablename__ = "secret_store_metadata"

    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.String(255), nullable=False)
    secret_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secrets.id'),
        index=True,
        nullable=False)

    secret = orm.relationship(
        "Secret", back_populates="secret_store_metadata")

    def __init__(self, key=None, value=None, check_exc=True):
        super(SecretStoreMetadatum, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for SecretStoreMetadatum entry.")

        if key is None and check_exc:
            raise exception.MissingArgumentError(msg.format("key"))
        self.key = key

        if value is None and check_exc:
            raise exception.MissingArgumentError(msg.format("value"))
        self.value = value

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {
            'key': self.key,
            'value': self.value
        }


class SecretUserMetadatum(BASE, SoftDeleteMixIn, ModelBase):
    """Represents Secret user metadatum for a single key-value pair."""

    __tablename__ = "secret_user_metadata"

    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.String(255), nullable=False)
    secret_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secrets.id'),
        index=True,
        nullable=False)

    secret = orm.relationship(
        "Secret",
        back_populates="secret_user_metadata")

    __table_args__ = (
        sa.UniqueConstraint('secret_id', 'key', name='_secret_key_uc'),
    )

    def __init__(self, key=None, value=None, check_exc=True):
        super(SecretUserMetadatum, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for SecretUserMetadatum entry.")

        if key is None and check_exc:
            raise exception.MissingArgumentError(msg.format("key"))
        self.key = key

        if value is None and check_exc:
            raise exception.MissingArgumentError(msg.format("value"))
        self.value = value

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {
            'key': self.key,
            'value': self.value
        }


class EncryptedDatum(BASE, SoftDeleteMixIn, ModelBase):
    """Represents the encrypted data for a Secret."""

    __tablename__ = 'encrypted_data'

    content_type = sa.Column(sa.String(255))
    secret_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secrets.id'),
        index=True,
        nullable=False)
    kek_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('kek_data.id'),
        index=True,
        nullable=False)
    # TODO(jwood) Why LargeBinary on Postgres (BYTEA) not work correctly?
    cypher_text = sa.Column(sa.Text)
    kek_meta_extended = sa.Column(sa.Text)

    # Eager load this relationship via 'lazy=False'.
    kek_meta_project = orm.relationship("KEKDatum", lazy=False)

    def __init__(self, secret=None, kek_datum=None, check_exc=True):
        """Creates encrypted datum from a secret and KEK metadata."""
        super(EncryptedDatum, self).__init__()

        if secret:
            self.secret_id = secret.id

        if kek_datum:
            self.kek_id = kek_datum.id
            self.kek_meta_project = kek_datum

        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'content_type': self.content_type}


class KEKDatum(BASE, SoftDeleteMixIn, ModelBase):
    """Key encryption key (KEK) metadata model.

    Represents the key encryption key (KEK) metadata associated with a process
    used to encrypt/decrypt secret information.

    When a secret is encrypted, in addition to the cypher text, the Barbican
    encryption process produces a KEK metadata object. The cypher text is
    stored via the EncryptedDatum model above, whereas the metadata is stored
    within this model. Decryption processes utilize this KEK metadata
    to decrypt the associated cypher text.

    Note that this model is intended to be agnostic to the specific means used
    to encrypt/decrypt the secret information, so please do not place vendor-
    specific attributes here.

    Note as well that each Project will have at most one 'active=True' KEKDatum
    instance at a time, representing the most recent KEK metadata instance
    to use for encryption processes performed on behalf of the Project.
    KEKDatum instances that are 'active=False' are associated to previously
    used encryption processes for the Project, that eventually should be
    rotated and deleted with the Project's active KEKDatum.
    """

    __tablename__ = 'kek_data'

    plugin_name = sa.Column(sa.String(255), nullable=False)
    kek_label = sa.Column(sa.String(255))
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id', name='kek_data_project_fk'),
        index=True,
        nullable=False)
    active = sa.Column(sa.Boolean, nullable=False, default=True)
    bind_completed = sa.Column(sa.Boolean, nullable=False, default=False)
    algorithm = sa.Column(sa.String(255))
    bit_length = sa.Column(sa.Integer)
    mode = sa.Column(sa.String(255))
    plugin_meta = sa.Column(sa.Text)

    project = orm.relationship('Project', back_populates='keks')

    def __index__(self, check_exc=True):
        super(KEKDatum, self).__init__()

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'algorithm': self.algorithm}


class Order(BASE, SoftDeleteMixIn, ModelBase):
    """Represents an Order in the datastore.

    Orders are requests for Barbican to generate secrets,
    ranging from symmetric, asymmetric keys to automated
    requests to Certificate Authorities to generate SSL
    certificates.
    """

    __tablename__ = 'orders'

    type = sa.Column(sa.String(255), nullable=False, default='key')
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id', name='orders_project_fk'),
        index=True,
        nullable=False)
    error_status_code = sa.Column(sa.String(16))
    error_reason = sa.Column(sa.String(ERROR_REASON_LENGTH))
    meta = sa.Column(JsonBlob(), nullable=True)
    secret_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secrets.id'),
        index=True,
        nullable=True)
    container_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('containers.id'),
        index=True,
        nullable=True)
    sub_status = sa.Column(sa.String(SUB_STATUS_LENGTH), nullable=True)
    sub_status_message = sa.Column(
        sa.String(SUB_STATUS_MESSAGE_LENGTH),
        nullable=True)
    creator_id = sa.Column(sa.String(255))

    project = orm.relationship('Project', back_populates='orders')
    order_plugin_metadata = orm.relationship(
        "OrderPluginMetadatum",
        collection_class=col.attribute_mapped_collection('key'),
        back_populates="order",
        cascade="all, delete-orphan",
        cascade_backrefs=False)
    order_barbican_metadata = orm.relationship(
        "OrderBarbicanMetadatum",
        collection_class=col.attribute_mapped_collection('key'),
        back_populates="order",
        cascade="all, delete-orphan",
        cascade_backrefs=False)

    def __init__(self, parsed_request=None, check_exc=True):
        """Creates a Order entity from a dict."""
        super(Order, self).__init__()
        if parsed_request:
            self.type = parsed_request.get('type')
            self.meta = parsed_request.get('meta')
            self.status = States.ACTIVE
            self.sub_status = parsed_request.get('sub_status')
            self.sub_status_message = parsed_request.get(
                'sub_status_message')
            self.creator_id = parsed_request.get('creator_id')

    def set_error_reason_safely(self, error_reason_raw):
        """Ensure error reason does not raise database attribute exceptions."""
        self.error_reason = error_reason_raw[:ERROR_REASON_LENGTH]

    def set_sub_status_safely(self, sub_status_raw):
        """Ensure sub-status does not raise database attribute exceptions."""
        self.sub_status = sub_status_raw[:SUB_STATUS_LENGTH]

    def set_sub_status_message_safely(self, sub_status_message_raw):
        """Ensure status message doesn't raise database attrib. exceptions."""
        self.sub_status_message = sub_status_message_raw[
            :SUB_STATUS_MESSAGE_LENGTH
        ]

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        for k, v in self.order_plugin_metadata.items():
            v.delete(session)
        for k, v in self.order_barbican_metadata.items():
            v.delete(session)

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        ret = {
            'type': self.type,
            'meta': self.meta,
            'order_id': self.id
        }
        if self.secret_id:
            ret['secret_id'] = self.secret_id
        if self.container_id:
            ret['container_id'] = self.container_id
        if self.error_status_code:
            ret['error_status_code'] = self.error_status_code
        if self.error_reason:
            ret['error_reason'] = self.error_reason
        if self.sub_status:
            ret['sub_status'] = self.sub_status
        if self.sub_status_message:
            ret['sub_status_message'] = self.sub_status_message
        if self.creator_id:
            ret['creator_id'] = self.creator_id
        return ret


class OrderPluginMetadatum(BASE, SoftDeleteMixIn, ModelBase):
    """Represents Order plugin metadatum for a single key-value pair.

    This entity is used to store plugin-specific metadata on behalf of an
    Order instance.
    """

    __tablename__ = "order_plugin_metadata"

    order_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('orders.id'),
        index=True,
        nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.String(255), nullable=False)

    order = orm.relationship("Order", back_populates="order_plugin_metadata")

    def __init__(self, key=None, value=None, check_exc=True):
        super(OrderPluginMetadatum, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for OrderPluginMetadatum entry.")

        if key is None and check_exc:
            raise exception.MissingArgumentError(msg.format("key"))
        self.key = key

        if value is None and check_exc:
            raise exception.MissingArgumentError(msg.format("value"))
        self.value = value

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'key': self.key,
                'value': self.value}


class OrderBarbicanMetadatum(BASE, SoftDeleteMixIn, ModelBase):
    """Represents Order barbican metadatum for a single key-value pair.

    This entity is used to store barbican-specific metadata on behalf of an
    Order instance.  This is data that is stored by the server to help
    process the order through its life cycle, but which is not in the original
    request.
    """

    __tablename__ = "order_barbican_metadata"

    order_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('orders.id'),
        index=True,
        nullable=False)
    key = sa.Column(sa.String(255), nullable=False)
    value = sa.Column(sa.Text, nullable=False)

    order = orm.relationship("Order", back_populates="order_barbican_metadata")

    def __init__(self, key=None, value=None, check_exc=True):
        super(OrderBarbicanMetadatum, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for OrderBarbicanMetadatum entry.")

        if key is None and check_exc:
            raise exception.MissingArgumentError(msg.format("key"))
        self.key = key

        if value is None and check_exc:
            raise exception.MissingArgumentError(msg.format("value"))
        self.value = value

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'key': self.key,
                'value': self.value}


class OrderRetryTask(BASE, SoftDeleteMixIn, ModelBase):

    __tablename__ = "order_retry_tasks"
    __table_args__ = {"mysql_engine": "InnoDB"}
    __table_initialized__ = False

    id = sa.Column(
        sa.String(36),
        primary_key=True,
        default=utils.generate_uuid)
    order_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("orders.id"),
        index=True,
        nullable=False)
    retry_task = sa.Column(sa.Text, nullable=False)
    retry_at = sa.Column(sa.DateTime, default=None, nullable=False)
    retry_args = sa.Column(JsonBlob(), nullable=False)
    retry_kwargs = sa.Column(JsonBlob(), nullable=False)
    retry_count = sa.Column(sa.Integer, nullable=False, default=0)

    def __index__(self, check_exc):
        super(OrderRetryTask, self).__init__()


class Container(BASE, SoftDeleteMixIn, ModelBase):
    """Represents a Container for Secrets in the datastore.

    Containers store secret references. Containers are owned by Projects.
    Containers can be generic or have a predefined type. Predefined typed
    containers allow users to store structured key relationship
    inside Barbican.
    """

    __tablename__ = 'containers'

    name = sa.Column(sa.String(255))
    type = sa.Column(
        sa.Enum(
            'generic', 'rsa', 'dsa', 'certificate',
            name='container_types'))
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id', name='containers_project_fk'),
        index=True,
        nullable=False)
    creator_id = sa.Column(sa.String(255))

    project = orm.relationship('Project', back_populates='containers')
    consumers = orm.relationship('ContainerConsumerMetadatum')
    container_acls = orm.relationship(
        'ContainerACL',
        back_populates='container',
        lazy=False)
    container_secrets = orm.relationship(
        "ContainerSecret",
        primaryjoin='and_(ContainerSecret.container_id==Container.id, ContainerSecret.deleted!=True)',  # noqa: E501
        back_populates='container',
        lazy=False,
    )

    def __init__(self, parsed_request=None, check_exc=True):
        """Creates a Container entity from a dict."""
        super(Container, self).__init__()

        if parsed_request:
            self.name = parsed_request.get('name')
            self.type = parsed_request.get('type')
            self.status = States.ACTIVE
            self.creator_id = parsed_request.get('creator_id')

            secret_refs = parsed_request.get('secret_refs')
            if secret_refs:
                for secret_ref in parsed_request.get('secret_refs'):
                    container_secret = ContainerSecret()
                    container_secret.name = secret_ref.get('name')
                    # TODO(hgedikli) move this into a common location
                    # TODO(hgedikli) validate provided url
                    # TODO(hgedikli) parse out secret_id with regex
                    secret_id = secret_ref.get('secret_ref')
                    if secret_id.endswith('/'):
                        secret_id = secret_id.rsplit('/', 2)[1]
                    elif '/' in secret_id:
                        secret_id = secret_id.rsplit('/', 1)[1]
                    else:
                        secret_id = secret_id
                    container_secret.secret_id = secret_id
                    self.container_secrets.append(container_secret)

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        for container_secret in self.container_secrets:
            session.delete(container_secret)

        for container_acl in self.container_acls:
            session.delete(container_acl)

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'container_id': self.id,
                'name': self.name,
                'type': self.type,
                'creator_id': self.creator_id,
                'secret_refs': [
                    {
                        'secret_id': container_secret.secret_id,
                        'name': container_secret.name
                        if hasattr(container_secret, 'name') else None
                    } for container_secret in self.container_secrets],
                'consumers': [
                    {
                        'name': consumer.name,
                        'URL': consumer.URL
                    } for consumer in self.consumers if not consumer.deleted
                ]}


class ContainerConsumerMetadatum(BASE, SoftDeleteMixIn, ModelBase):
    """Stores Consumer Registrations for Containers in the datastore.

    Services can register interest in Containers. Services will provide a type
    and a URL for the object that is using the Container.
    """

    __tablename__ = 'container_consumer_metadata'

    container_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('containers.id'),
        index=True,
        nullable=False)
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id'),
        index=True,
        nullable=True)
    name = sa.Column(sa.String(36))
    URL = sa.Column(sa.String(255))
    data_hash = sa.Column(sa.CHAR(64))

    __table_args__ = (
        sa.UniqueConstraint('data_hash',
                            name='_consumer_hashed_container_name_url_uc'),
        sa.Index('values_index', 'container_id', 'name', 'URL')
    )

    def __init__(self, container_id=None, project_id=None,
                 parsed_request=None, check_exc=True):
        """Registers a Consumer to a Container."""
        super(ContainerConsumerMetadatum, self).__init__()

        # TODO(john-wood-w) This class should really be immutable due to the
        # data_hash attribute.
        if container_id and parsed_request:
            self.container_id = container_id
            self.project_id = project_id
            self.name = parsed_request.get('name')
            self.URL = parsed_request.get('URL')
            hash_text = ''.join((self.container_id, self.name, self.URL))
            self.data_hash = hashlib.sha256(hash_text.
                                            encode('utf-8')).hexdigest()
            self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'name': self.name,
                'URL': self.URL}


class TransportKey(BASE, SoftDeleteMixIn, ModelBase):
    """Transport Key model for wrapping secrets in transit

    Represents the transport key used for wrapping secrets in transit
    to/from clients when storing/retrieving secrets.
    """

    __tablename__ = 'transport_keys'

    plugin_name = sa.Column(sa.String(255), nullable=False)
    transport_key = sa.Column(sa.Text, nullable=False)

    def __init__(self, plugin_name=None, transport_key=None, check_exc=True):
        """Creates transport key entity."""
        super(TransportKey, self).__init__()

        msg = u._("Must supply non-None {0} argument for TransportKey entry.")

        if plugin_name is None and check_exc:
            raise exception.MissingArgumentError(msg.format("plugin_name"))

        self.plugin_name = plugin_name

        if transport_key is None and check_exc:
            raise exception.MissingArgumentError(msg.format("transport_key"))

        self.transport_key = transport_key
        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'transport_key_id': self.id,
                'plugin_name': self.plugin_name}


class CertificateAuthority(BASE, ModelBase):
    """CertificateAuthority model to specify the CAs available to Barbican

    Represents the CAs available for certificate issuance to Barbican.
    """

    __tablename__ = 'certificate_authorities'

    plugin_name = sa.Column(sa.String(255), nullable=False)
    plugin_ca_id = sa.Column(sa.Text, nullable=False)
    expiration = sa.Column(sa.DateTime, default=None)
    creator_id = sa.Column(sa.String(255), nullable=True)
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id', name='cas_project_fk'),
        nullable=True)

    ca_meta = orm.relationship(
        'CertificateAuthorityMetadatum',
        collection_class=col.attribute_mapped_collection('key'),
        back_populates='ca',
        cascade="all, delete-orphan",
        cascade_backrefs=False)
    project_cas = orm.relationship(
        'ProjectCertificateAuthority',
        back_populates='ca')
    preferred_ca = orm.relationship(
        'PreferredCertificateAuthority',
        back_populates='ca')

    def __init__(self, parsed_ca_in=None, check_exc=True):
        """Creates certificate authority entity."""
        super(CertificateAuthority, self).__init__()

        msg = u._("Must supply Non-None {0} argument "
                  "for CertificateAuthority entry.")

        parsed_ca = dict(parsed_ca_in)

        plugin_name = parsed_ca.pop('plugin_name', None)
        if plugin_name is None:
            raise exception.MissingArgumentError(msg.format("plugin_name"))
        self.plugin_name = plugin_name

        plugin_ca_id = parsed_ca.pop('plugin_ca_id', None)
        if plugin_ca_id is None:
            raise exception.MissingArgumentError(msg.format("plugin_ca_id"))
        self.plugin_ca_id = plugin_ca_id

        expiration = parsed_ca.pop('expiration', None)
        self.expiration = self._iso_to_datetime(expiration)

        creator_id = parsed_ca.pop('creator_id', None)
        if creator_id is not None:
            self.creator_id = creator_id

        project_id = parsed_ca.pop('project_id', None)
        if project_id is not None:
            self.project_id = project_id

        for key in parsed_ca:
            meta = CertificateAuthorityMetadatum(key, parsed_ca[key])
            self.ca_meta[key] = meta

        self.status = States.ACTIVE

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        for k, v in self.ca_meta.items():
            v.delete(session)

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        if self.expiration:
            expiration = self.expiration.isoformat()
        else:
            expiration = None

        return {
            'ca_id': self.id,
            'plugin_name': self.plugin_name,
            'plugin_ca_id': self.plugin_ca_id,
            'expiration': expiration,
            'meta': [
                {
                    meta['key']: meta['value']
                } for key, meta in self.ca_meta.items()
            ]
        }


class CertificateAuthorityMetadatum(BASE, ModelBase):
    """Represents CA metadatum for a single key-value pair."""

    __tablename__ = "certificate_authority_metadata"

    key = sa.Column(sa.String(255), index=True, nullable=False)
    value = sa.Column(sa.Text, nullable=False)
    ca_id = sa.Column(
        sa.String(36), sa.ForeignKey('certificate_authorities.id'),
        index=True, nullable=False)

    ca = orm.relationship('CertificateAuthority', back_populates='ca_meta')

    __table_args__ = (
        sa.UniqueConstraint(
            'ca_id', 'key', name='_certificate_authority_metadatum_uc',
        ),
    )

    def __init__(self, key=None, value=None, check_exc=True):
        super(CertificateAuthorityMetadatum, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for CertificateAuthorityMetadatum entry.")

        if key is None and check_exc:
            raise exception.MissingArgumentError(msg.format("key"))
        self.key = key

        if value is None and check_exc:
            raise exception.MissingArgumentError(msg.format("value"))
        self.value = value

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {
            'key': self.key,
            'value': self.value
        }


class ProjectCertificateAuthority(BASE, ModelBase):
    """Stores CAs available for a project.

    Admins can define a set of CAs that are available for use in a particular
    project.  There can be multiple entries for any given project.
    """

    __tablename__ = 'project_certificate_authorities'

    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id'),
        index=True,
        nullable=False)
    ca_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('certificate_authorities.id'),
        index=True,
        nullable=False)

    project = orm.relationship('Project', back_populates='cas')
    ca = orm.relationship("CertificateAuthority", back_populates='project_cas')

    __table_args__ = (
        sa.UniqueConstraint(
            'project_id', 'ca_id', name='_project_certificate_authority_uc',
        ),
    )

    def __init__(self, project_id=None, ca_id=None, check_exc=True):
        """Registers a Consumer to a Container."""
        super(ProjectCertificateAuthority, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for ProjectCertificateAuthority entry.")

        if project_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("project_id"))
        self.project_id = project_id

        if ca_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("ca_id"))
        self.ca_id = ca_id

        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'project_id': self.project_id,
                'ca_id': self.ca_id}


class PreferredCertificateAuthority(BASE, ModelBase):
    """Stores preferred CAs for any project.

    Admins can define a set of CAs available for issuance requests for
    any project in the ProjectCertificateAuthority table..
    """

    __tablename__ = 'preferred_certificate_authorities'

    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id'),
        index=True,
        unique=True,
        nullable=False)
    ca_id = sa.Column(
        sa.String(36),
        sa.ForeignKey(
            'certificate_authorities.id',
            name='preferred_certificate_authorities_fk'),
        index=True,
        nullable=False)

    project = orm.relationship(
        'Project',
        back_populates='preferred_ca',
        uselist=False)
    ca = orm.relationship(
        'CertificateAuthority',
        back_populates='preferred_ca')

    def __init__(self, project_id=None, ca_id=None, check_exc=True):
        """Registers a Consumer to a Container."""
        super(PreferredCertificateAuthority, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for PreferredCertificateAuthority entry.")

        if project_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("project_id"))
        self.project_id = project_id

        if ca_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("ca_id"))
        self.ca_id = ca_id

        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'project_id': self.project_id,
                'ca_id': self.ca_id}


class SecretACL(BASE, ModelBase):
    """Stores Access Control List (ACL) for a secret.

    Class to define whitelist of user ids who are allowed specific operation
    on a secret. List of user ids is defined via SecretACLUser via
    acl_users association.
    Creator_only flag helps in making a secret private for
    non-admin project users who may have access otherwise.

    SecretACL deletes are not soft-deletes.
    """

    __tablename__ = 'secret_acls'

    secret_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secrets.id'),
        index=True,
        nullable=False)
    operation = sa.Column(sa.String(255), nullable=False)
    project_access = sa.Column(sa.Boolean, nullable=False, default=True)

    secret = orm.relationship(
        'Secret',
        back_populates='secret_acls',
        lazy=False)
    acl_users = orm.relationship(
        'SecretACLUser',
        back_populates='secret_acl',
        lazy=False,
        cascade="all, delete-orphan",
        cascade_backrefs=False)

    __table_args__ = (
        sa.UniqueConstraint(
            'secret_id', 'operation', name='_secret_acl_operation_uc',
        ),
    )

    def __init__(self, secret_id=None, operation=None, project_access=None,
                 user_ids=None, check_exc=True):
        """Creates secret ACL entity."""
        super(SecretACL, self).__init__()

        msg = u._("Must supply non-None {0} argument for SecretACL entry.")

        if secret_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("secret_id"))
        self.secret_id = secret_id

        if operation is None and check_exc:
            raise exception.MissingArgumentError(msg.format("operation"))
        self.operation = operation

        if project_access is not None:
            self.project_access = project_access
        self.status = States.ACTIVE
        if user_ids is not None and isinstance(user_ids, list):
            userids = set(user_ids)  # remove duplicate if any
            for user_id in userids:
                acl_user = SecretACLUser(self.id, user_id)
                self.acl_users.append(acl_user)

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        for acl_user in self.acl_users:
            acl_user.delete(session)

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields.

        Adds non-deleted acl related users from relationship if there.
        """
        users = [acl_user.user_id for acl_user in self.acl_users
                 if not acl_user.deleted]
        fields = {'acl_id': self.id,
                  'secret_id': self.secret_id,
                  'operation': self.operation,
                  'project_access': self.project_access}
        if users:
            fields['users'] = users
        return fields


class ContainerACL(BASE, ModelBase):
    """Stores Access Control List (ACL) for a container.

    Class to define whitelist of user ids who are allowed specific operation
    on a container. List of user ids is defined in ContainerACLUser via
    acl_users association.
    Creator_only flag helps in making a container private for
    non-admin project users who may have access otherwise.

    ContainerACL deletes are not soft-deletes.
    """

    __tablename__ = 'container_acls'

    container_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('containers.id'),
        index=True,
        nullable=False)
    operation = sa.Column(sa.String(255), nullable=False)
    project_access = sa.Column(sa.Boolean, nullable=False, default=True)

    container = orm.relationship(
        'Container',
        back_populates='container_acls',
        lazy=False)
    acl_users = orm.relationship(
        'ContainerACLUser',
        back_populates='container_acl',
        cascade="all, delete-orphan",
        cascade_backrefs=False)

    __table_args__ = (
        sa.UniqueConstraint(
            'container_id', 'operation', name='_container_acl_operation_uc',
        ),
    )

    def __init__(self, container_id=None, operation=None, project_access=None,
                 user_ids=None, check_exc=True):
        """Creates container ACL entity."""
        super(ContainerACL, self).__init__()

        msg = u._("Must supply non-None {0} argument for ContainerACL entry.")

        if container_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("container_id"))
        self.container_id = container_id

        if operation is None and check_exc:
            raise exception.MissingArgumentError(msg.format("operation"))
        self.operation = operation

        if project_access is not None:
            self.project_access = project_access
        self.status = States.ACTIVE

        if user_ids is not None and isinstance(user_ids, list):
            userids = set(user_ids)  # remove duplicate if any
            for user_id in userids:
                acl_user = ContainerACLUser(self.id, user_id)
                self.acl_users.append(acl_user)

    def _do_delete_children(self, session):
        """Sub-class hook: delete children relationships."""
        for acl_user in self.acl_users:
            acl_user.delete(session)

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields.

        Adds non-deleted acl related users from relationship if there.
        """
        users = [acl_user.user_id for acl_user in self.acl_users
                 if not acl_user.deleted]
        fields = {'acl_id': self.id,
                  'container_id': self.container_id,
                  'operation': self.operation,
                  'project_access': self.project_access}
        if users:
            fields['users'] = users
        return fields


class SecretACLUser(BASE, ModelBase):
    """Stores user id for a secret ACL.

    This class provides way to store list of users associated with a
    specific ACL operation.

    SecretACLUser deletes are not soft-deletes.
    """

    __tablename__ = 'secret_acl_users'

    acl_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secret_acls.id'),
        index=True,
        nullable=False)
    user_id = sa.Column(sa.String(255), nullable=False)

    secret_acl = orm.relationship(
        'SecretACL',
        back_populates='acl_users',
        lazy=False)

    __table_args__ = (
        sa.UniqueConstraint('acl_id', 'user_id', name='_secret_acl_user_uc'),
    )

    def __init__(self, acl_id=None, user_id=None, check_exc=True):
        """Creates secret ACL user entity."""
        super(SecretACLUser, self).__init__()

        msg = u._("Must supply non-None {0} argument for SecretACLUser entry.")

        self.acl_id = acl_id
        if user_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("user_id"))
        self.user_id = user_id
        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'acl_id': self.acl_id,
                'user_id': self.user_id}


class ContainerACLUser(BASE, ModelBase):
    """Stores user id for a container ACL.

    This class provides way to store list of users associated with a
    specific ACL operation.

    ContainerACLUser deletes are not soft-deletes.
    """

    __tablename__ = 'container_acl_users'

    acl_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('container_acls.id'),
        index=True,
        nullable=False)
    user_id = sa.Column(sa.String(255), nullable=False)

    container_acl = orm.relationship(
        'ContainerACL',
        back_populates='acl_users',
        lazy=False)

    __table_args__ = (
        sa.UniqueConstraint(
            'acl_id', 'user_id', name='_container_acl_user_uc',
        ),
    )

    def __init__(self, acl_id=None, user_id=None, check_exc=True):
        """Creates container ACL user entity."""
        super(ContainerACLUser, self).__init__()

        msg = u._("Must supply non-None {0} argument for ContainerACLUser "
                  "entry.")

        self.acl_id = acl_id
        if user_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("user_id"))
        self.user_id = user_id
        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'acl_id': self.acl_id,
                'user_id': self.user_id}


class ProjectQuotas(BASE, ModelBase):
    """Stores Project Quotas.

    Class to define project specific resource quotas.

    Project quota deletes are not soft-deletes.
    """

    __tablename__ = 'project_quotas'

    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id', name='project_quotas_fk'),
        index=True,
        nullable=False)
    secrets = sa.Column(sa.Integer, nullable=True)
    orders = sa.Column(sa.Integer, nullable=True)
    containers = sa.Column(sa.Integer, nullable=True)
    consumers = sa.Column(sa.Integer, nullable=True)
    cas = sa.Column(sa.Integer, nullable=True)

    project = orm.relationship('Project', back_populates='project_quotas')

    def __init__(self, project_id=None, parsed_project_quotas=None,
                 check_exc=True):
        """Creates Project Quotas entity from a project and a dict.

        :param project_id: the internal id of the project with quotas
        :param parsed_project_quotas: a dict with the keys matching the
        resources for which quotas are to be set, and the values containing
        the quota value to be set for this project and that resource.
        :return: None
        """
        super(ProjectQuotas, self).__init__()

        msg = u._("Must supply non-None {0} argument for ProjectQuotas entry.")

        if project_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("project_id"))
        self.project_id = project_id

        if parsed_project_quotas is None:
            self.secrets = None
            self.orders = None
            self.containers = None
            self.consumers = None
            self.cas = None
        else:
            self.secrets = parsed_project_quotas.get('secrets')
            self.orders = parsed_project_quotas.get('orders')
            self.containers = parsed_project_quotas.get('containers')
            self.consumers = parsed_project_quotas.get('consumers')
            self.cas = parsed_project_quotas.get('cas')

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        ret = {
            'project_id': self.project_id,
        }
        if self.secrets:
            ret['secrets'] = self.secrets
        if self.orders:
            ret['orders'] = self.orders
        if self.containers:
            ret['containers'] = self.containers
        if self.consumers:
            ret['consumers'] = self.consumers
        if self.cas:
            ret['cas'] = self.cas
        return ret


class SecretStores(BASE, ModelBase):
    """List of secret stores defined via service configuration.

    This class provides a list of secret stores entities with their respective
    secret store plugin and crypto plugin names.

    SecretStores deletes are NOT soft-deletes.
    """

    __tablename__ = 'secret_stores'

    store_plugin = sa.Column(sa.String(255), nullable=False)
    crypto_plugin = sa.Column(sa.String(255), nullable=True)
    global_default = sa.Column(sa.Boolean, nullable=False, default=False)
    name = sa.Column(sa.String(255), nullable=False)

    project_store = orm.relationship(
        'ProjectSecretStore',
        back_populates='secret_store')

    __table_args__ = (
        sa.UniqueConstraint(
            'store_plugin', 'crypto_plugin',
            name='_secret_stores_plugin_names_uc'),
        sa.UniqueConstraint('name', name='_secret_stores_name_uc'),
    )

    def __init__(self, name=None, store_plugin=None, crypto_plugin=None,
                 global_default=None, check_exc=True):
        """Creates secret store entity."""
        super(SecretStores, self).__init__()

        msg = u._("Must supply non-Blank {0} argument for SecretStores entry.")

        if not name and check_exc:
            raise exception.MissingArgumentError(msg.format("name"))
        if not store_plugin and check_exc:
            raise exception.MissingArgumentError(msg.format("store_plugin"))

        self.store_plugin = store_plugin
        self.name = name
        self.crypto_plugin = crypto_plugin
        if global_default is not None:
            self.global_default = global_default

        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'secret_store_id': self.id,
                'store_plugin': self.store_plugin,
                'crypto_plugin': self.crypto_plugin,
                'global_default': self.global_default,
                'name': self.name}


class ProjectSecretStore(BASE, ModelBase):
    """Stores secret store to be used for new project secrets.

    This class maintains secret store and project mapping so that new project
    secret entries uses it as plugin backend.

    ProjectSecretStores deletes are NOT soft-deletes.
    """

    __tablename__ = 'project_secret_store'

    secret_store_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('secret_stores.id'),
        index=True,
        nullable=False)
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey('projects.id'),
        index=True,
        nullable=False)

    secret_store = orm.relationship(
        'SecretStores',
        back_populates='project_store')
    project = orm.relationship(
        'Project',
        back_populates='preferred_secret_store')

    __table_args__ = (
        sa.UniqueConstraint(
            'project_id', name='_project_secret_store_project_uc'),
    )

    def __init__(self, project_id=None, secret_store_id=None, check_exc=True):
        """Creates project secret store mapping entity."""
        super(ProjectSecretStore, self).__init__()

        msg = u._("Must supply non-None {0} argument for ProjectSecretStore "
                  " entry.")

        if not project_id and check_exc:
            raise exception.MissingArgumentError(msg.format("project_id"))
        self.project_id = project_id
        if not secret_store_id and check_exc:
            raise exception.MissingArgumentError(msg.format("secret_store_id"))
        self.secret_store_id = secret_store_id

        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {'secret_store_id': self.secret_store_id,
                'project_id': self.project_id}


class SecretConsumerMetadatum(BASE, SoftDeleteMixIn, ModelBase):
    """Stores Consumer Registrations for Secrets in the datastore.

    Services can register interest in Secrets. Services will provide a
    resource type and a resource id for the object that is using the Secret.
    """

    __tablename__ = "secret_consumer_metadata"

    secret_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("secrets.id"),
        index=True,
        nullable=False
    )
    project_id = sa.Column(
        sa.String(36),
        sa.ForeignKey("projects.id"),
        index=True,
        nullable=True
    )
    service = sa.Column(sa.String(255), nullable=False)
    resource_type = sa.Column(sa.String(255), nullable=False)
    resource_id = sa.Column(sa.String(36), index=True, nullable=False)

    secret = orm.relationship("Secret", back_populates="consumers")

    __table_args__ = (
        sa.UniqueConstraint(
            "secret_id", "service", "resource_type", "resource_id",
            name="_secret_consumer_resource_uc"
        ),
    )

    def __init__(self, secret_id=None, project_id=None, service=None,
                 resource_type=None, resource_id=None, check_exc=True):
        """Registers a Consumer to a Secret."""
        super(SecretConsumerMetadatum, self).__init__()

        msg = u._("Must supply non-None {0} argument "
                  "for SecretConsumerMetadatum entry.")

        if secret_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("secret_id"))
        if project_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("project_id"))
        if service is None and check_exc:
            raise exception.MissingArgumentError(msg.format("service"))
        if resource_type is None and check_exc:
            raise exception.MissingArgumentError(msg.format("resource_type"))
        if resource_id is None and check_exc:
            raise exception.MissingArgumentError(msg.format("resource_id"))

        self.secret_id = secret_id
        self.project_id = project_id
        self.service = service
        self.resource_type = resource_type
        self.resource_id = resource_id
        self.status = States.ACTIVE

    def _do_extra_dict_fields(self):
        """Sub-class hook method: return dict of fields."""
        return {
            "service": self.service,
            "resource_type": self.resource_type,
            "resource_id": self.resource_id,
        }



# Copyright 2022 OpenStack Foundation
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Update secret consumers unique constraint to mach the updated spec

Revision ID: 8c74e2d7f1ff
Revises: 0f8c192a061f
Create Date: 2022-09-12 13:03:26.428642

"""

# revision identifiers, used by Alembic.
revision = '8c74e2d7f1ff'
down_revision = '0f8c192a061f'

from alembic import op
import sqlalchemy as sa


def upgrade():
    with op.batch_alter_table('secret_consumer_metadata') as batch_op:
        batch_op.alter_column('project_id',
                              existing_type=sa.VARCHAR(length=36),
                              nullable=True)
        batch_op.drop_constraint(
            '_secret_consumer_resource_uc', type_='unique')
        batch_op.create_unique_constraint(
            constraint_name='_secret_consumer_resource_uc',
            columns=['secret_id', 'service', 'resource_type', 'resource_id'])
        batch_op.create_index(
            index_name=op.f('ix_secret_consumer_metadata_project_id'),
            columns=['project_id'],
            unique=False)
        batch_op.create_foreign_key(constraint_name=op.f('fk_project_id'),
                                    referent_table='projects',
                                    local_cols=['project_id'],
                                    remote_cols=['id'])



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime

import fixtures
from oslo_utils import timeutils
import testtools

from barbican.common import exception
from barbican.model import models
from barbican.model import repositories
from barbican.plugin.interface import secret_store as ss
from barbican.tests import database_utils
from barbican.tests import fixture
from barbican.tests import utils


@utils.parameterized_test_case
class WhenTestingSecretRepository(database_utils.RepositoryTestCase):

    dataset_for_filter_tests = {
        'query_by_name': {
            'secret_1_dict': dict(name="name1"),
            'secret_2_dict': dict(name="name2"),
            'query_dict': dict(name="name1")
        },
        'query_by_algorithm': {
            'secret_1_dict': dict(algorithm="algorithm1"),
            'secret_2_dict': dict(algorithm="algorithm2"),
            'query_dict': dict(alg="algorithm1")
        },
        'query_by_mode': {
            'secret_1_dict': dict(mode="mode1"),
            'secret_2_dict': dict(mode="mode2"),
            'query_dict': dict(mode="mode1")
        },
        'query_by_bit_length': {
            'secret_1_dict': dict(bit_length=1024),
            'secret_2_dict': dict(bit_length=2048),
            'query_dict': dict(bits=1024)
        },
        'query_by_secret_type': {
            'secret_1_dict': dict(secret_type=ss.SecretType.SYMMETRIC),
            'secret_2_dict': dict(secret_type=ss.SecretType.OPAQUE),
            'query_dict': dict(secret_type=ss.SecretType.SYMMETRIC)
        },
    }

    def setUp(self):
        super(WhenTestingSecretRepository, self).setUp()
        self.repo = repositories.SecretRepo()

    def test_get_secret_list(self):
        session = self.repo.get_session()

        project = models.Project()
        project.external_id = "my keystone id"
        project.save(session=session)

        secret_model = models.Secret()
        secret_model.project_id = project.id
        secret = self.repo.create_from(secret_model, session=session)

        session.commit()

        secrets, offset, limit, total = self.repo.get_secret_list(
            "my keystone id",
            session=session,
        )

        self.assertEqual([secret.id], [s.id for s in secrets])
        self.assertEqual(0, offset)
        self.assertEqual(10, limit)
        self.assertEqual(1, total)

    def test_get_secret_by_id(self):
        session = self.repo.get_session()

        project = models.Project()
        project.external_id = "my keystone id"
        project.save(session=session)

        secret_model = models.Secret()
        secret_model.project_id = project.id
        secret = self.repo.create_from(secret_model, session=session)

        session.commit()

        db_secret = self.repo.get_secret_by_id(secret.id)
        self.assertIsNotNone(db_secret)

    def test_should_raise_notfound_exception(self):
        self.assertRaises(exception.NotFound, self.repo.get_secret_by_id,
                          "invalid_id", suppress_exception=False)

    def test_should_suppress_notfound_exception(self):
        self.assertIsNone(self.repo.get_secret_by_id("invalid_id",
                                                     suppress_exception=True))

    @utils.parameterized_dataset(dataset_for_filter_tests)
    def test_get_secret_list_with_filter(self, secret_1_dict, secret_2_dict,
                                         query_dict):
        session = self.repo.get_session()

        project = models.Project()
        project.external_id = "my keystone id"
        project.save(session=session)

        secret_1_dict['project_id'] = project.id
        secret1 = self.repo.create_from(
            models.Secret(secret_1_dict),
            session=session,
        )
        secret_2_dict['project_id'] = project.id
        secret2 = self.repo.create_from(
            models.Secret(secret_2_dict),
            session=session,
        )

        session.commit()

        secrets, offset, limit, total = self.repo.get_secret_list(
            "my keystone id",
            session=session,
            **query_dict
        )
        resulting_secret_ids = [s.id for s in secrets]
        self.assertIn(secret1.id, resulting_secret_ids)
        self.assertNotIn(secret2.id, resulting_secret_ids)
        self.assertEqual(0, offset)
        self.assertEqual(10, limit)
        self.assertEqual(1, total)

    def test_get_by_create_date_nothing(self):
        session = self.repo.get_session()
        secrets, offset, limit, total = self.repo.get_secret_list(
            "my keystone id",
            bits=1024,
            session=session,
            suppress_exception=True
        )

        self.assertEqual([], secrets)
        self.assertEqual(0, offset)
        self.assertEqual(10, limit)
        self.assertEqual(0, total)

    def test_do_entity_name(self):
        self.assertEqual("Secret", self.repo._do_entity_name())

    def test_should_raise_no_result_found(self):
        session = self.repo.get_session()

        self.assertRaises(
            exception.NotFound,
            self.repo.get_secret_list,
            "my keystone id",
            session=session,
            suppress_exception=False)

    def test_should_get_count_zero(self):
        session = self.repo.get_session()

        project = models.Project()
        project.external_id = "my keystone id"
        project.save(session=session)

        session.commit()
        count = self.repo.get_count(project.id, session=session)

        self.assertEqual(0, count)

    def test_should_get_count_one(self):
        session = self.repo.get_session()

        project = models.Project()
        project.external_id = "my keystone id"
        project.save(session=session)

        secret_model = models.Secret()
        secret_model.project_id = project.id
        self.repo.create_from(secret_model, session=session)

        session.commit()
        count = self.repo.get_count(project.id, session=session)

        self.assertEqual(1, count)

    def test_should_get_count_one_after_delete(self):
        session = self.repo.get_session()

        project = models.Project()
        project.external_id = "my keystone id"
        project.save(session=session)

        secret_model = models.Secret()
        secret_model.project_id = project.id
        self.repo.create_from(secret_model, session=session)

        secret_model = models.Secret()
        secret_model.project_id = project.id
        self.repo.create_from(secret_model, session=session)

        session.commit()
        count = self.repo.get_count(project.id, session=session)
        self.assertEqual(2, count)

        self.repo.delete_entity_by_id(secret_model.id, "my keystone id",
                                      session=session)
        session.commit()

        count = self.repo.get_count(project.id, session=session)
        self.assertEqual(1, count)

    def test_should_get_count_one_after_expiration(self):
        current_time = timeutils.utcnow()
        tomorrow = current_time + datetime.timedelta(days=1)
        yesterday = current_time - datetime.timedelta(days=1)

        session = self.repo.get_session()

        project = models.Project()
        project.external_id = "my keystone id"
        project.save(session=session)

        secret_model = models.Secret()
        secret_model.project_id = project.id
        secret_model.expiration = tomorrow
        self.repo.create_from(secret_model, session=session)

        secret_model = models.Secret()
        secret_model.project_id = project.id
        secret_model.expiration = yesterday
        self.repo.create_from(secret_model, session=session)

        session.commit()

        count = self.repo.get_count(project.id, session=session)
        self.assertEqual(1, count)


class WhenTestingQueryFilters(testtools.TestCase,
                              fixtures.TestWithFixtures):

    def setUp(self):
        super(WhenTestingQueryFilters, self).setUp()
        self._session_fixture = self.useFixture(fixture.SessionQueryFixture())
        self.session = self._session_fixture.Session()
        self.query = self.session.query(models.Secret)
        self.repo = repositories.SecretRepo()

    def test_data_includes_six_secrets(self):
        self.assertEqual(6, len(self.query.all()))

    def test_sort_by_name_defaults_ascending(self):
        query = self.repo._build_sort_filter_query(self.query, 'name')
        secrets = query.all()
        self.assertEqual('A', secrets[0].name)

    def test_sort_by_name_desc(self):
        query = self.repo._build_sort_filter_query(self.query, 'name:desc')
        secrets = query.all()
        self.assertEqual('F', secrets[0].name)

    def test_sort_by_created_asc(self):
        query = self.repo._build_sort_filter_query(self.query, 'created:asc')
        secrets = query.all()
        self.assertEqual('A', secrets[0].name)

    def test_sort_by_updated_desc(self):
        query = self.repo._build_sort_filter_query(self.query, 'updated:desc')
        secrets = query.all()
        self.assertEqual('F', secrets[0].name)

    def test_filter_by_created_on_new_years(self):
        query = self.repo._build_date_filter_query(
            self.query, 'created_at',
            '2016-01-01T00:00:00'
        )
        secrets = query.all()
        self.assertEqual(1, len(secrets))
        self.assertEqual('A', secrets[0].name)

    def test_filter_by_created_after_march(self):
        query = self.repo._build_date_filter_query(
            self.query, 'created_at',
            'gt:2016-03-01T00:00:00'
        )
        secrets = query.all()
        self.assertEqual(3, len(secrets))

    def test_filter_by_created_on_or_after_march(self):
        query = self.repo._build_date_filter_query(
            self.query, 'created_at',
            'gte:2016-03-01T00:00:00'
        )
        secrets = query.all()
        self.assertEqual(4, len(secrets))

    def test_filter_by_created_before_march(self):
        query = self.repo._build_date_filter_query(
            self.query, 'created_at',
            'lt:2016-03-01T00:00:00'
        )
        secrets = query.all()
        self.assertEqual(2, len(secrets))

    def test_filter_by_created_on_or_before_march(self):
        query = self.repo._build_date_filter_query(
            self.query, 'created_at',
            'lte:2016-03-01T00:00:00'
        )
        secrets = query.all()
        self.assertEqual(3, len(secrets))

    def test_filter_by_created_between_march_and_may_inclusive(self):
        query = self.repo._build_date_filter_query(
            self.query, 'created_at',
            'gte:2016-03-01T00:00:00,lte:2016-05-01T00:00:00'
        )
        secrets = query.all()
        secret_names = [s.name for s in secrets]

        self.assertEqual(3, len(secrets))
        self.assertIn('C', secret_names)
        self.assertIn('D', secret_names)
        self.assertIn('E', secret_names)



# Copyright  2011-2012 OpenStack LLC.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import oslo_context

from barbican.common import policy


class RequestContext(oslo_context.context.RequestContext):
    """User security context object

    Stores information about the security context under which the user
    accesses the system, as well as additional request information.
    """

    def __init__(self, policy_enforcer=None, **kwargs):
        # prefer usage of 'project' instead of 'tenant'
        if policy_enforcer:
            self.policy_enforcer = policy_enforcer
        else:
            policy.init()
            self.policy_enforcer = policy.get_enforcer()
        super(RequestContext, self).__init__(**kwargs)

    def to_dict(self):
        out_dict = super(RequestContext, self).to_dict()
        out_dict['roles'] = self.roles

        return out_dict

    @classmethod
    def from_dict(cls, values):
        return cls(**values)



# Copyright (c) 2016 IBM
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import datetime
from unittest import mock

from oslo_db import exception as db_exc
from oslo_utils import timeutils
from sqlalchemy.exc import IntegrityError

from barbican.model import clean
from barbican.model import models
from barbican.model import repositories as repos
from barbican.tests import database_utils as utils


def _create_project(project_name):
    """Wrapper to create a project and clean"""
    def project_decorator(test_func):
        def project_wrapper(self, *args, **kwargs):
            project = utils.create_project(external_id=project_name)
            kwargs['project'] = project
            test_result = test_func(self, *args, **kwargs)
            project.delete()
            return test_result
        return project_wrapper
    return project_decorator


def _entry_exists(entry):
    """Check to see if entry should exist in the database"""
    model = entry.__class__
    entry_id = entry.id
    session = repos.get_session()
    query = session.query(model).filter(model.id == entry_id)
    count = query.count()
    return count >= 1


def _entry_is_soft_deleted(entry):
    model = entry.__class__
    entry_id = entry.id
    session = repos.get_session()
    query = session.query(model)
    result = query.filter(model.id == entry_id).first().deleted
    return result


def _setup_entry(name, *args, **kwargs):
    func_name = "create_" + name
    if not hasattr(utils, func_name):
        raise Exception("Cannot create an entry called %s", name)
    func = getattr(utils, func_name)
    kwargs['session'] = repos.get_session()
    entry = func(*args, **kwargs)
    return entry


class WhenTestingDBCleanUpCommand(utils.RepositoryTestCase):

    def tearDown(self):
        super(WhenTestingDBCleanUpCommand, self).tearDown()
        repos.rollback()

    @_create_project("my keystone id")
    def test_soft_deleted_secret_orders(self, project):
        """Test that secrets without child order get deleted"""
        # Create a secret tied to an order and one secret that is not
        secret1 = _setup_entry('secret', project=project)
        secret2 = _setup_entry('secret', project=project)
        order = _setup_entry('order', project=project, secret=secret1)

        # Delete secrets
        secret1.delete()
        secret2.delete()
        clean.cleanup_parent_with_no_child(models.Secret, models.Order)

        # Assert that only secret2 is removed
        self.assertTrue(_entry_exists(secret1))
        self.assertFalse(_entry_exists(secret2))

        # delete order and secret
        order.delete()
        clean.cleanup_all()

        self.assertFalse(_entry_exists(order))
        self.assertFalse(_entry_exists(secret2))

    def test_cleanup_soft_deletes_transport_keys(self):
        """Test Cleaning up soft deleted transport keys"""
        # create transport key
        transport_key = _setup_entry('transport_key')

        # delete transport key
        transport_key.delete()
        clean.cleanup_all()
        self.assertFalse(_entry_exists(transport_key))

    @_create_project("my keystone id")
    def test_cleanup_soft_deletes_secrets(self, project):
        """Test cleaning up secrets and secret_meta"""
        # create secret and secret_meta
        secret = _setup_entry('secret', project=project)
        secret_metadatum = _setup_entry('secret_metadatum', secret=secret)
        secret_user_metadatum = _setup_entry('secret_user_metadatum',
                                             secret=secret)
        kek_datum = _setup_entry('kek_datum', project=project)
        enc_datum = _setup_entry('encrypted_datum', secret=secret,
                                 kek_datum=kek_datum)
        # delete secret, it should automatically delete
        # secret_metadatum, enc_datum, and secret_user_metadatum
        # kek_datum should still exist
        secret.delete()
        clean.cleanup_all()
        self.assertFalse(_entry_exists(secret))
        self.assertFalse(_entry_exists(secret_metadatum))
        self.assertFalse(_entry_exists(secret_user_metadatum))
        self.assertFalse(_entry_exists(enc_datum))
        self.assertTrue(_entry_exists(kek_datum))

    @_create_project("my keystone id")
    def test_cleanup_soft_deletes_containers(self, project):
        """Test cleaning up containers and secrets"""
        # create container, secret, and container_secret
        container = _setup_entry('container', project=project)
        secret = _setup_entry('secret', project=project)
        container_secret = _setup_entry('container_secret',
                                        container=container, secret=secret)

        # delete container secret and container
        container.delete()
        clean.cleanup_all()

        # check that container secret and container are deleted
        # but secret still exists
        self.assertFalse(_entry_exists(container_secret))
        self.assertFalse(_entry_exists(container))
        self.assertTrue(_entry_exists(secret))

        # cleanup secrets
        secret.delete()
        clean.cleanup_all()
        self.assertFalse(_entry_exists(secret))

    @_create_project("my keystone id")
    def test_cleanup_container_with_order_child(self, project):
        container = _setup_entry('container', project=project)
        secret = _setup_entry('secret', project=project)
        secret_container = _setup_entry('container_secret',
                                        container=container, secret=secret)
        order = _setup_entry('order', project=project, secret=secret,
                             container=container)

        container.delete()
        clean.cleanup_all()

        # only the secret_container should be removed from the database
        # since it is a child of the container
        self.assertFalse(_entry_exists(secret_container))
        self.assertTrue(_entry_exists(secret))
        self.assertTrue(_entry_exists(order))
        # container should still exist since child order still exists
        self.assertTrue(_entry_exists(container))

        order.delete()
        clean.cleanup_all()

        # assert that only the secret exists
        self.assertFalse(_entry_exists(order))
        self.assertFalse(_entry_exists(container))
        self.assertTrue(_entry_exists(secret))

        secret.delete()
        clean.cleanup_all()
        # the secret should now be able to be removed
        self.assertFalse(_entry_exists(secret))

    @_create_project("my clean order keystone id")
    def test_cleanup_orders(self, project):
        """Test cleaning up an order and it's children"""
        # create order, order meta, and plugin meta, and retry task
        order = _setup_entry('order', project=project)
        order_barbican_meta_data = _setup_entry('order_meta_datum',
                                                order=order)
        order_plugin_metadata = _setup_entry('order_plugin_metadatum',
                                             order=order)
        order_retry_task = _setup_entry('order_retry', order=order)

        # soft delete order and retry task,
        # it should automatically delete the children
        order.delete()
        order_retry_task.delete()
        clean.cleanup_all()

        # assert everything has been cleaned up
        self.assertFalse(_entry_exists(order))
        self.assertFalse(_entry_exists(order_plugin_metadata))
        self.assertFalse(_entry_exists(order_retry_task))
        self.assertFalse(_entry_exists(order_barbican_meta_data))

    @_create_project("my clean order with child keystone id")
    def test_cleanup_order_with_child(self, project):
        """Test cleaning up an order with a child"""
        # create order and retry task
        order = _setup_entry('order', project=project)
        order_retry_task = _setup_entry('order_retry', order=order)

        # soft delete order and retry task,
        #  it should automatically delete the children
        order.delete()
        clean.cleanup_all()

        # assert that the order was not cleaned due to child
        self.assertTrue(_entry_exists(order))
        self.assertTrue(_entry_exists(order_retry_task))

        order_retry_task.delete()
        clean.cleanup_all()

        # assert everything has been cleaned up
        self.assertFalse(_entry_exists(order))
        self.assertFalse(_entry_exists(order_retry_task))

    @_create_project("my keystone id")
    def test_cleanup_soft_deletion_date(self, project):
        """Test cleaning up entries within date"""
        secret = _setup_entry('secret', project=project)
        order = order = _setup_entry('order', project=project, secret=secret)
        current_time = timeutils.utcnow()
        tomorrow = current_time + datetime.timedelta(days=1)
        yesterday = current_time - datetime.timedelta(days=1)
        secret.delete()
        order.delete()

        # Assert that nothing is deleted due to date
        clean.cleanup_softdeletes(models.Order, threshold_date=yesterday)
        clean.cleanup_parent_with_no_child(models.Secret, models.Order,
                                           threshold_date=yesterday)
        self.assertTrue(_entry_exists(secret))
        self.assertTrue(_entry_exists(order))

        # Assert that everything is deleted due to date
        clean.cleanup_softdeletes(models.Order, threshold_date=tomorrow)
        clean.cleanup_parent_with_no_child(models.Secret, models.Order,
                                           threshold_date=tomorrow)
        self.assertFalse(_entry_exists(secret))
        self.assertFalse(_entry_exists(order))

    @_create_project("my keystone id")
    def test_soft_deleting_expired_secrets(self, project):
        """Test soft deleting secrets that are expired"""

        current_time = timeutils.utcnow()
        tomorrow = current_time + datetime.timedelta(days=1)
        yesterday = current_time - datetime.timedelta(days=1)

        not_expired_secret = _setup_entry('secret', project=project)
        expired_secret = _setup_entry('secret', project=project)
        not_expired_secret.expiration = tomorrow
        expired_secret.expiration = yesterday

        # Create children for expired secret
        expired_secret_store_metadatum = _setup_entry('secret_metadatum',
                                                      secret=expired_secret)
        expired_secret_user_metadatum = _setup_entry('secret_user_metadatum',
                                                     secret=expired_secret)
        kek_datum = _setup_entry('kek_datum', project=project)
        expired_enc_datum = _setup_entry('encrypted_datum',
                                         secret=expired_secret,
                                         kek_datum=kek_datum)
        container = _setup_entry('container', project=project)
        expired_container_secret = _setup_entry('container_secret',
                                                container=container,
                                                secret=expired_secret)
        expired_acl_secret = _setup_entry('acl_secret',
                                          secret=expired_secret,
                                          user_ids=["fern", "chris"])

        clean.soft_delete_expired_secrets(current_time)
        self.assertTrue(_entry_is_soft_deleted(expired_secret))
        self.assertFalse(_entry_is_soft_deleted(not_expired_secret))

        # Make sure the children of the expired secret are soft deleted as well
        self.assertTrue(_entry_is_soft_deleted(expired_enc_datum))
        self.assertTrue(_entry_is_soft_deleted(expired_container_secret))
        self.assertTrue(_entry_is_soft_deleted(expired_secret_store_metadatum))
        self.assertTrue(_entry_is_soft_deleted(expired_secret_user_metadatum))
        self.assertFalse(_entry_exists(expired_acl_secret))

    def test_cleaning_unassociated_projects(self):
        """Test cleaning projects that have no child entries"""
        childless_project = _setup_entry('project',
                                         external_id="childless project")
        project_with_children = _setup_entry(
            'project',
            external_id="project with children")

        project_children_list = list()
        project_children_list.append(
            _setup_entry('kek_datum', project=project_with_children))

        project_children_list.append(
            _setup_entry('secret', project=project_with_children))

        container = _setup_entry('container', project=project_with_children)
        project_children_list.append(container)
        project_children_list.append(
            _setup_entry('container_consumer_meta', container=container))
        cert_authority = _setup_entry('certificate_authority',
                                      project=project_with_children)
        project_children_list.append(cert_authority)
        project_children_list.append(
            _setup_entry('preferred_cert_authority',
                         cert_authority=cert_authority))

        project_children_list.append(
            _setup_entry('project_cert_authority',
                         certificate_authority=cert_authority))
        project_children_list.append(_setup_entry('project_quotas',
                                     project=project_with_children))

        clean.cleanup_unassociated_projects()
        self.assertTrue(_entry_exists(project_with_children))
        self.assertFalse(_entry_exists(childless_project))

        container.delete()
        for child in project_children_list:
            child.delete()
        clean.cleanup_all()
        clean.cleanup_unassociated_projects()
        self.assertFalse(_entry_exists(project_with_children))

    @mock.patch('barbican.model.clean.cleanup_all')
    @mock.patch('barbican.model.clean.soft_delete_expired_secrets')
    @mock.patch('barbican.model.clean.cleanup_unassociated_projects')
    @mock.patch('barbican.model.clean.repo')
    @mock.patch('barbican.model.clean.log')
    @mock.patch('barbican.model.clean.CONF')
    def test_clean_up_command(self, mock_conf, mock_log, mock_repo,
                              mock_clean_unc_projects,
                              mock_soft_del_expire_secrets, mock_clean_all):
        """Tests the clean command"""
        test_sql_url = "mysql+pymysql://notrealuser:datab@127.0.0.1/barbican't"
        min_num_days = 91
        do_clean_unassociated_projects = True
        do_soft_delete_expired_secrets = True
        verbose = True
        test_log_file = "/tmp/sometempfile"
        clean.clean_command(test_sql_url, min_num_days,
                            do_clean_unassociated_projects,
                            do_soft_delete_expired_secrets, verbose,
                            test_log_file)
        set_calls = [mock.call('debug', True),
                     mock.call('log_file', test_log_file),
                     mock.call('connection', test_sql_url, 'database')]
        mock_conf.set_override.assert_has_calls(set_calls)

        clear_calls = [mock.call('debug'), mock.call('log_file'),
                       mock.call('connection', 'database')]
        mock_conf.clear_override.assert_has_calls(clear_calls)
        self.assertTrue(mock_repo.setup_database_engine_and_factory.called)
        self.assertTrue(mock_repo.commit.called)
        self.assertTrue(mock_repo.clear.called)
        self.assertTrue(mock_clean_unc_projects.called)
        self.assertTrue(mock_soft_del_expire_secrets)
        self.assertTrue(mock_clean_all)

    @mock.patch('barbican.model.clean.cleanup_all')
    @mock.patch('barbican.model.clean.soft_delete_expired_secrets')
    @mock.patch('barbican.model.clean.cleanup_unassociated_projects')
    @mock.patch('barbican.model.clean.repo')
    @mock.patch('barbican.model.clean.log')
    @mock.patch('barbican.model.clean.CONF')
    def test_clean_up_command_with_false_args(
        self, mock_conf, mock_log, mock_repo, mock_clean_unc_projects,
            mock_soft_del_expire_secrets, mock_clean_all):
        """Tests the clean command with false args"""
        test_sql_url = None
        min_num_days = -1
        do_clean_unassociated_projects = False
        do_soft_delete_expired_secrets = False
        verbose = None
        test_log_file = None
        clean.clean_command(test_sql_url, min_num_days,
                            do_clean_unassociated_projects,
                            do_soft_delete_expired_secrets, verbose,
                            test_log_file)
        mock_conf.set_override.assert_not_called()
        mock_conf.clear_override.assert_not_called()
        self.assertTrue(mock_repo.setup_database_engine_and_factory.called)
        self.assertTrue(mock_repo.commit.called)
        self.assertTrue(mock_repo.clear.called)
        self.assertTrue(mock_clean_all)
        self.assertFalse(mock_clean_unc_projects.called)
        self.assertFalse(mock_soft_del_expire_secrets.called)

    @mock.patch('barbican.model.clean.cleanup_all',
                side_effect=IntegrityError("", "", "", ""))
    @mock.patch('barbican.model.clean.repo')
    @mock.patch('barbican.model.clean.log')
    @mock.patch('barbican.model.clean.CONF')
    def test_clean_up_command_with_exception(
            self, mock_conf, mock_log, mock_repo, mock_clean_all):
        """Tests that the clean command throws exceptions"""
        args = ("sql", 2, False, False, False, "/tmp/nope")
        self.assertRaises(IntegrityError, clean.clean_command, *args)
        self.assertTrue(mock_repo.rollback.called)

    @_create_project("my integrity error keystone id")
    def test_db_cleanup_raise_integrity_error(self, project):
        """Test that an integrity error is thrown

        This test tests the invalid scenario where
        the secret meta was not marked for deletion during the secret deletion.
        We want to make sure an integrity error is thrown during clean up.
        """
        # create secret
        secret = _setup_entry('secret', project=project)
        secret_metadatum = _setup_entry('secret_metadatum', secret=secret)

        # delete parent but not child and assert integrity error
        secret.deleted = True
        secret_metadatum.deleted = False

        self.assertRaises(db_exc.DBReferenceError, clean.cleanup_all)



# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from barbican.common import exception
from barbican.common import utils
from barbican.model import models
from barbican.model import repositories as repos
from barbican.plugin.interface import secret_store
from barbican.plugin import store_crypto
from barbican.plugin.util import translations as tr


def _get_transport_key_model(key_spec, transport_key_needed, project_id):
    key_model = None
    if transport_key_needed:
        # get_plugin_store() will throw an exception if no suitable
        # plugin with transport key is found
        plugin_manager = secret_store.get_manager()
        store_plugin = plugin_manager.get_plugin_store(
            key_spec=key_spec, transport_key_needed=True,
            project_id=project_id)
        plugin_name = utils.generate_fullname_for(store_plugin)

        key_repo = repos.get_transport_key_repository()
        key_model = key_repo.get_latest_transport_key(plugin_name)

        if not key_model or not store_plugin.is_transport_key_current(
                key_model.transport_key):
            # transport key does not exist or is not current.
            # need to get a new transport key
            transport_key = store_plugin.get_transport_key()
            new_key_model = models.TransportKey(plugin_name, transport_key)
            key_model = key_repo.create_from(new_key_model)
    return key_model


def _get_plugin_name_and_transport_key(transport_key_id):
    plugin_name = None
    transport_key = None
    if transport_key_id is not None:
        transport_key_repo = repos.get_transport_key_repository()
        try:
            transport_key_model = transport_key_repo.get(
                entity_id=transport_key_id)
        except exception.NotFound:
            raise exception.ProvidedTransportKeyNotFound(str(transport_key_id))

        plugin_name = transport_key_model.plugin_name
        if plugin_name is None:
            raise ValueError("Invalid plugin name for transport key")

        transport_key = transport_key_model.transport_key

    return plugin_name, transport_key


def store_secret(unencrypted_raw, content_type_raw, content_encoding,
                 secret_model, project_model,
                 transport_key_needed=False,
                 transport_key_id=None):
    """Store a provided secret into secure backend."""
    if _secret_already_has_stored_data(secret_model):
        raise ValueError('Secret already has encrypted data stored for it.')

    # Create a KeySpec to find a plugin that will support storing the secret
    key_spec = secret_store.KeySpec(alg=secret_model.algorithm,
                                    bit_length=secret_model.bit_length,
                                    mode=secret_model.mode)

    # If there is no secret data to store, then just create Secret entity and
    #   leave. A subsequent call to this method should provide both the Secret
    #   entity created here *and* the secret data to store into it.
    if not unencrypted_raw:
        key_model = _get_transport_key_model(key_spec, transport_key_needed,
                                             project_id=project_model.id)

        _save_secret_in_repo(secret_model, project_model)
        return secret_model, key_model

    plugin_name, transport_key = _get_plugin_name_and_transport_key(
        transport_key_id)

    unencrypted, content_type = tr.normalize_before_encryption(
        unencrypted_raw, content_type_raw, content_encoding,
        secret_model.secret_type, enforce_text_only=True)

    plugin_manager = secret_store.get_manager()
    store_plugin = plugin_manager.get_plugin_store(key_spec=key_spec,
                                                   plugin_name=plugin_name,
                                                   project_id=project_model.id)

    secret_dto = secret_store.SecretDTO(type=secret_model.secret_type,
                                        secret=unencrypted,
                                        key_spec=key_spec,
                                        content_type=content_type,
                                        transport_key=transport_key)

    secret_metadata = _store_secret_using_plugin(store_plugin, secret_dto,
                                                 secret_model, project_model)
    _save_secret_in_repo(secret_model, project_model)
    _save_secret_metadata_in_repo(secret_model, secret_metadata, store_plugin,
                                  content_type)

    return secret_model, None


def get_secret(requesting_content_type, secret_model, project_model,
               twsk=None, transport_key=None):
    secret_metadata = _get_secret_meta(secret_model)

    # NOTE: */* is the pecan default meaning no content type sent in.  In this
    # case we should use the mime type stored in the metadata.
    if requesting_content_type == '*/*':
        requesting_content_type = secret_metadata['content_type']

    tr.analyze_before_decryption(requesting_content_type)

    if twsk is not None:
        secret_metadata['trans_wrapped_session_key'] = twsk
        secret_metadata['transport_key'] = transport_key

    # Locate a suitable plugin to store the secret.
    plugin_manager = secret_store.get_manager()
    retrieve_plugin = plugin_manager.get_plugin_retrieve_delete(
        secret_metadata.get('plugin_name'))

    # Retrieve the secret.
    secret_dto = _get_secret(
        retrieve_plugin, secret_metadata, secret_model, project_model)

    if twsk is not None:
        del secret_metadata['transport_key']
        del secret_metadata['trans_wrapped_session_key']

    # Denormalize the secret.
    return tr.denormalize_after_decryption(secret_dto.secret,
                                           requesting_content_type)


def get_transport_key_id_for_retrieval(secret_model):
    """Return a transport key ID for retrieval if the plugin supports it."""

    secret_metadata = _get_secret_meta(secret_model)

    plugin_manager = secret_store.get_manager()
    retrieve_plugin = plugin_manager.get_plugin_retrieve_delete(
        secret_metadata.get('plugin_name'))

    transport_key_id = retrieve_plugin.get_transport_key()
    return transport_key_id


def generate_secret(spec, content_type, project_model):
    """Generate a secret and store into a secure backend."""

    # Locate a suitable plugin to store the secret.
    key_spec = secret_store.KeySpec(alg=spec.get('algorithm'),
                                    bit_length=spec.get('bit_length'),
                                    mode=spec.get('mode'))

    plugin_manager = secret_store.get_manager()
    generate_plugin = plugin_manager.get_plugin_generate(
        key_spec, project_id=project_model.id)

    # Create secret model to eventually save metadata to.
    secret_model = models.Secret(spec)
    secret_model['secret_type'] = secret_store.SecretType.SYMMETRIC

    # Generate the secret.
    secret_metadata = _generate_symmetric_key(
        generate_plugin, key_spec, secret_model, project_model, content_type)

    # Save secret and metadata.
    _save_secret_in_repo(secret_model, project_model)
    _save_secret_metadata_in_repo(secret_model, secret_metadata,
                                  generate_plugin, content_type)

    return secret_model


def generate_asymmetric_secret(spec, content_type, project_model):
    """Generate an asymmetric secret and store into a secure backend."""
    # Locate a suitable plugin to store the secret.
    key_spec = secret_store.KeySpec(alg=spec.get('algorithm'),
                                    bit_length=spec.get('bit_length'),
                                    passphrase=spec.get('passphrase'))

    plugin_manager = secret_store.get_manager()
    generate_plugin = plugin_manager.get_plugin_generate(
        key_spec, project_id=project_model.id)

    # Create secret models to eventually save metadata to.
    private_secret_model = models.Secret(spec)
    private_secret_model['secret_type'] = secret_store.SecretType.PRIVATE
    public_secret_model = models.Secret(spec)
    public_secret_model['secret_type'] = secret_store.SecretType.PUBLIC
    passphrase_secret_model = (models.Secret(spec)
                               if spec.get('passphrase') else None)
    if passphrase_secret_model:
        passphrase_type = secret_store.SecretType.PASSPHRASE
        passphrase_secret_model['secret_type'] = passphrase_type

    asymmetric_meta_dto = _generate_asymmetric_key(
        generate_plugin,
        key_spec,
        private_secret_model,
        public_secret_model,
        passphrase_secret_model,
        project_model,
        content_type
    )

    _save_secret_in_repo(private_secret_model, project_model)
    _save_secret_metadata_in_repo(private_secret_model,
                                  asymmetric_meta_dto.private_key_meta,
                                  generate_plugin,
                                  content_type)

    _save_secret_in_repo(public_secret_model, project_model)
    _save_secret_metadata_in_repo(public_secret_model,
                                  asymmetric_meta_dto.public_key_meta,
                                  generate_plugin,
                                  content_type)

    if passphrase_secret_model:
        _save_secret_in_repo(passphrase_secret_model, project_model)
        _save_secret_metadata_in_repo(passphrase_secret_model,
                                      asymmetric_meta_dto.passphrase_meta,
                                      generate_plugin,
                                      content_type)

    container_model = _create_container_for_asymmetric_secret(spec,
                                                              project_model)
    _save_asymmetric_secret_in_repo(
        container_model, private_secret_model, public_secret_model,
        passphrase_secret_model)

    return container_model


def delete_secret(secret_model, project_id):
    """Remove a secret from secure backend."""

    secret_metadata = _get_secret_meta(secret_model)

    # We should only try to delete a secret using the plugin interface if
    # there's the metadata available. This addresses bug/1377330.
    if secret_metadata:
        # Locate a suitable plugin to delete the secret from.
        plugin_manager = secret_store.get_manager()
        delete_plugin = plugin_manager.get_plugin_retrieve_delete(
            secret_metadata.get('plugin_name'))

        # Delete the secret from plugin storage.
        delete_plugin.delete_secret(secret_metadata)

    # Delete the secret from data model.
    secret_repo = repos.get_secret_repository()
    secret_repo.delete_entity_by_id(entity_id=secret_model.id,
                                    external_project_id=project_id)


def _store_secret_using_plugin(store_plugin, secret_dto, secret_model,
                               project_model):
    if isinstance(store_plugin, store_crypto.StoreCryptoAdapterPlugin):
        context = store_crypto.StoreCryptoContext(
            project_model,
            secret_model=secret_model)
        secret_metadata = store_plugin.store_secret(secret_dto, context)
    else:
        secret_metadata = store_plugin.store_secret(secret_dto)
    return secret_metadata


def _generate_symmetric_key(
        generate_plugin, key_spec, secret_model, project_model, content_type):
    if isinstance(generate_plugin, store_crypto.StoreCryptoAdapterPlugin):
        context = store_crypto.StoreCryptoContext(
            project_model,
            secret_model=secret_model,
            content_type=content_type)
        secret_metadata = generate_plugin.generate_symmetric_key(
            key_spec, context)
    else:
        secret_metadata = generate_plugin.generate_symmetric_key(key_spec)
    return secret_metadata


def _generate_asymmetric_key(generate_plugin, key_spec, private_secret_model,
                             public_secret_model, passphrase_secret_model,
                             project_model, content_type):
    if isinstance(generate_plugin, store_crypto.StoreCryptoAdapterPlugin):
        context = store_crypto.StoreCryptoContext(
            project_model,
            private_secret_model=private_secret_model,
            public_secret_model=public_secret_model,
            passphrase_secret_model=passphrase_secret_model,
            content_type=content_type)
        asymmetric_meta_dto = generate_plugin.generate_asymmetric_key(
            key_spec, context)
    else:
        asymmetric_meta_dto = generate_plugin.generate_asymmetric_key(key_spec)
    return asymmetric_meta_dto


def _get_secret(retrieve_plugin, secret_metadata, secret_model, project_model):
    if isinstance(retrieve_plugin, store_crypto.StoreCryptoAdapterPlugin):
        context = store_crypto.StoreCryptoContext(
            project_model,
            secret_model=secret_model)
        secret_dto = retrieve_plugin.get_secret(secret_model.secret_type,
                                                secret_metadata,
                                                context)
    else:
        secret_dto = retrieve_plugin.get_secret(secret_model.secret_type,
                                                secret_metadata)
    return secret_dto


def _get_secret_meta(secret_model):
    if secret_model:
        secret_meta_repo = repos.get_secret_meta_repository()
        return secret_meta_repo.get_metadata_for_secret(secret_model.id)
    else:
        return {}


def _save_secret_metadata_in_repo(secret_model, secret_metadata,
                                  store_plugin, content_type):
    """Add secret metadata to a secret."""

    if not secret_metadata:
        secret_metadata = {}

    secret_metadata['plugin_name'] = utils.generate_fullname_for(store_plugin)
    secret_metadata['content_type'] = content_type

    secret_meta_repo = repos.get_secret_meta_repository()
    secret_meta_repo.save(secret_metadata, secret_model)


def _save_secret_in_repo(secret_model, project_model):
    """Save a Secret entity."""

    secret_repo = repos.get_secret_repository()
    # Create Secret entities in data store.
    if not secret_model.id:
        secret_model.project_id = project_model.id
        secret_repo.create_from(secret_model)
    else:
        secret_repo.save(secret_model)


def _secret_already_has_stored_data(secret_model):
    if not secret_model:
        return False
    return secret_model.encrypted_data or secret_model.secret_store_metadata


def _create_container_for_asymmetric_secret(spec, project_model):
    container_model = models.Container()
    container_model.name = spec.get('name')
    container_model.type = spec.get('algorithm', '').lower()
    container_model.status = models.States.ACTIVE
    container_model.project_id = project_model.id
    container_model.creator_id = spec.get('creator_id')
    return container_model


def _save_asymmetric_secret_in_repo(container_model, private_secret_model,
                                    public_secret_model,
                                    passphrase_secret_model):
    container_repo = repos.get_container_repository()
    container_repo.create_from(container_model)

    # create container_secret for private_key
    _create_container_secret_association('private_key',
                                         private_secret_model,
                                         container_model)

    # create container_secret for public_key
    _create_container_secret_association('public_key',
                                         public_secret_model,
                                         container_model)

    if passphrase_secret_model:
        # create container_secret for passphrase
        _create_container_secret_association('private_key_passphrase',
                                             passphrase_secret_model,
                                             container_model)


def _create_container_secret_association(assoc_name, secret_model,
                                         container_model):
    container_secret = models.ContainerSecret()
    container_secret.name = assoc_name
    container_secret.container_id = container_model.id
    container_secret.secret_id = secret_model.id

    container_secret_repo = repos.get_container_secret_repository()
    container_secret_repo.create_from(container_secret)



# Copyright (c) 2013-2014 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Barbican exception subclasses
"""

from barbican import i18n as u

_FATAL_EXCEPTION_FORMAT_ERRORS = False


class BarbicanException(Exception):
    """Base Barbican Exception

    To correctly use this class, inherit from it and define
    a 'message' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """
    message = u._("An unknown exception occurred")

    def __init__(self, message_arg=None, *args, **kwargs):
        if not message_arg:
            message_arg = self.message
        try:
            self.message = message_arg % kwargs
        except Exception as e:
            if _FATAL_EXCEPTION_FORMAT_ERRORS:
                raise e
            else:
                # at least get the core message out if something happened
                pass
        super(BarbicanException, self).__init__(self.message)


class BarbicanHTTPException(BarbicanException):
    """Base Barbican Exception to handle HTTP responses

    To correctly use this class, inherit from it and define the following
    properties:

    - message: The message that will be displayed in the server log.
    - client_message: The message that will actually be outputted to the
                      client.
    - status_code: The HTTP status code that should be returned.
                   The default status code is 500.
    """
    client_message = u._("failure seen - please contact site administrator.")
    status_code = 500

    def __init__(self, message_arg=None, client_message=None, *args, **kwargs):
        if not client_message:
            client_message = self.client_message
        try:
            self.client_message = client_message % kwargs
        except Exception as e:
            if _FATAL_EXCEPTION_FORMAT_ERRORS:
                raise e
            else:
                # at least get the core message out if something happened
                pass
        super(BarbicanHTTPException, self).__init__(
            message_arg, self.client_message, *args, **kwargs)


class MissingArgumentError(BarbicanException):
    message = u._("Missing required argument.")


class MissingMetadataField(BarbicanHTTPException):
    message = u._("Missing required metadata field for %(required)s")
    client_message = message
    status_code = 400


class InvalidMetadataRequest(BarbicanHTTPException):
    message = u._("Invalid Metadata. Keys and Values must be Strings.")
    client_message = message
    status_code = 400


class InvalidMetadataKey(BarbicanHTTPException):
    message = u._("Invalid Key. Key must be URL safe.")
    client_message = message
    status_code = 400


class InvalidSubjectDN(BarbicanHTTPException):
    message = u._("Invalid subject DN: %(subject_dn)s")
    client_message = message
    status_code = 400


class InvalidContainer(BarbicanHTTPException):
    message = u._("Invalid container: %(reason)s")
    client_message = message
    status_code = 400


class InvalidExtensionsData(BarbicanHTTPException):
    message = u._("Invalid extensions data.")
    client_message = message
    status_code = 400


class InvalidCMCData(BarbicanHTTPException):
    message = u._("Invalid CMC Data")
    client_message = message
    status_code = 400


class InvalidPKCS10Data(BarbicanHTTPException):
    message = u._("Invalid PKCS10 Data: %(reason)s")
    client_message = message
    status_code = 400


class InvalidCertificateRequestType(BarbicanHTTPException):
    message = u._("Invalid Certificate Request Type")
    client_message = message
    status_code = 400


class CertificateExtensionsNotSupported(BarbicanHTTPException):
    message = u._("Extensions are not yet supported.  "
                  "Specify a valid profile instead.")
    client_message = message
    status_code = 400


class FullCMCNotSupported(BarbicanHTTPException):
    message = u._("Full CMC Requests are not yet supported.")
    client_message = message
    status_code = 400


class NotFound(BarbicanException):
    message = u._("An object with the specified identifier was not found.")


class ConstraintCheck(BarbicanException):
    message = u._("A defined SQL constraint check failed: %(error)s")


class NotSupported(BarbicanException):
    message = u._("Operation is not supported.")


class Invalid(BarbicanException):
    message = u._("Data supplied was not valid.")


class NoDataToProcess(BarbicanHTTPException):
    message = u._("No data supplied to process.")
    client_message = message
    status_code = 400


class LimitExceeded(BarbicanHTTPException):
    message = u._("The request returned a 413 Request Entity Too Large. This "
                  "generally means that rate limiting or a quota threshold "
                  "was breached.")
    client_message = u._("Provided information too large to process")
    status_code = 413

    def __init__(self, *args, **kwargs):
        super(LimitExceeded, self).__init__(*args, **kwargs)
        self.retry_after = (int(kwargs['retry']) if kwargs.get('retry')
                            else None)


class InvalidObject(BarbicanHTTPException):
    status_code = 400

    def __init__(self, *args, **kwargs):
        self.invalid_property = kwargs.get('property')
        self.message = u._("Failed to validate JSON information: ")
        self.client_message = u._("Provided object does not match "
                                  "schema '{schema}': "
                                  "{reason}. Invalid property: "
                                  "'{property}'").format(*args, **kwargs)
        self.message = self.message + self.client_message
        super(InvalidObject, self).__init__(*args, **kwargs)


class PayloadDecodingError(BarbicanHTTPException):
    status_code = 400
    message = u._("Error while attempting to decode payload.")
    client_message = u._("Unable to decode request data.")


class UnsupportedField(BarbicanHTTPException):
    message = u._("No support for value set on field '%(field)s' on "
                  "schema '%(schema)s': %(reason)s")
    client_message = u._("Provided field value is not supported")
    status_code = 400

    def __init__(self, *args, **kwargs):
        super(UnsupportedField, self).__init__(*args, **kwargs)
        self.invalid_field = kwargs.get('field')


class FeatureNotImplemented(BarbicanException):
    message = u._("Feature not implemented for value set on field "
                  "'%(field)s' on " "schema '%(schema)s': %(reason)s")

    def __init__(self, *args, **kwargs):
        super(FeatureNotImplemented, self).__init__(*args, **kwargs)
        self.invalid_field = kwargs.get('field')


class StoredKeyContainerNotFound(BarbicanException):
    message = u._("Container %(container_id)s does not exist for stored "
                  "key certificate generation.")


class StoredKeyPrivateKeyNotFound(BarbicanException):
    message = u._("Container %(container_id)s does not reference a private "
                  "key needed for stored key certificate generation.")


class ProvidedTransportKeyNotFound(BarbicanHTTPException):
    message = u._("Provided Transport key %(transport_key_id)s "
                  "could not be found")
    client_message = u._("Provided transport key was not found.")
    status_code = 400


class InvalidCAID(BarbicanHTTPException):
    message = u._("Invalid CA_ID: %(ca_id)s")
    client_message = u._("The ca_id provided in the request is invalid")
    status_code = 400


class CANotDefinedForProject(BarbicanHTTPException):
    message = u._("CA specified by ca_id %(ca_id)s not defined for project: "
                  "%(project_id)s")
    client_message = u._("The ca_id provided in the request is not defined "
                         "for this project")
    status_code = 403


class QuotaReached(BarbicanHTTPException):
    message = u._("Quota reached for project %(external_project_id)s. Only "
                  "%(quota)s %(resource_type)s are allowed.")
    client_message = u._("Creation not allowed because a quota has "
                         "been reached")
    status_code = 403

    def __init__(self, *args, **kwargs):
        super(QuotaReached, self).__init__(*args, **kwargs)
        self.external_project_id = kwargs.get('external_project_id')
        self.quota = kwargs.get('quota')
        self.resource_type = kwargs.get('resource_type')


class InvalidParentCA(BarbicanHTTPException):
    message = u._("Invalid Parent CA: %(parent_ca_ref)s")
    client_message = message
    status_code = 400


class SubCAsNotSupported(BarbicanHTTPException):
    message = u._("Plugin does not support generation of subordinate CAs")
    client_message = message
    status_code = 400


class SubCANotCreated(BarbicanHTTPException):
    message = u._("Errors in creating subordinate CA: %(name)")
    client_message = message


class CannotDeleteBaseCA(BarbicanHTTPException):
    message = u._("Only subordinate CAs can be deleted.")
    status_code = 403


class UnauthorizedSubCA(BarbicanHTTPException):
    message = u._("Subordinate CA is not owned by this project")
    client_message = message
    status_code = 403


class CannotDeletePreferredCA(BarbicanHTTPException):
    message = u._("A new project preferred CA must be set "
                  "before this one can be deleted.")
    status_code = 409


class BadSubCACreationRequest(BarbicanHTTPException):
    message = u._("Errors returned by CA when attempting to "
                  "create subordinate CA: %(reason)s")
    client_message = message
    status_code = 400


class SubCACreationErrors(BarbicanHTTPException):
    message = u._("Errors returned by CA when attempting to create "
                  "subordinate CA: %(reason)s")
    client_message = message


class SubCADeletionErrors(BarbicanHTTPException):
    message = u._("Errors returned by CA when attempting to delete "
                  "subordinate CA: %(reason)s")
    client_message = message


class PKCS11Exception(BarbicanException):
    message = u._("There was an error with the PKCS#11 library.")


class P11CryptoPluginKeyException(PKCS11Exception):
    message = u._("More than one key found for label")


class P11CryptoPluginException(PKCS11Exception):
    message = u._("General exception")


class P11CryptoKeyHandleException(PKCS11Exception):
    message = u._("No key handle was found")


class P11CryptoTokenException(PKCS11Exception):
    message = u._("No token was found in slot %(slot_id)s")


class TrustwayProteccioException(PKCS11Exception):
    message = u._("Trustway Proteccio HSM Error")


class MultipleStorePreferredPluginMissing(BarbicanException):
    """Raised when a preferred plugin is missing in service configuration."""
    def __init__(self, store_name):
        super(MultipleStorePreferredPluginMissing, self).__init__(
            u._("Preferred Secret Store plugin '{store_name}' is not "
                "currently set in service configuration. This is probably a "
                "server misconfiguration.").format(
                store_name=store_name)
        )
        self.store_name = store_name


class MultipleStorePluginStillInUse(BarbicanException):
    """Raised when a used plugin is missing in service configuration."""
    def __init__(self, store_name):
        super(MultipleStorePluginStillInUse, self).__init__(
            u._("Secret Store plugin '{store_name}' is still in use and can "
                "not be removed. Its missing in service configuration. This is"
                " probably a server misconfiguration.").format(
                store_name=store_name)
        )
        self.store_name = store_name


class MultipleSecretStoreLookupFailed(BarbicanException):
    """Raised when a plugin lookup suffix is missing during config read."""
    def __init__(self):
        msg = u._("Plugin lookup property 'stores_lookup_suffix' is not "
                  "defined in service configuration")
        super(MultipleSecretStoreLookupFailed, self).__init__(msg)


class MultipleStoreIncorrectGlobalDefault(BarbicanException):
    """Raised when a global default for only one plugin is not set to True."""
    def __init__(self, occurrence):
        msg = None
        if occurrence > 1:
            msg = u._("There are {count} plugins with global default as "
                      "True in service configuration. Only one plugin can have"
                      " this as True").format(count=occurrence)
        else:
            msg = u._("There is no plugin defined with global default as True."
                      " One of plugin must be identified as global default")

        super(MultipleStoreIncorrectGlobalDefault, self).__init__(msg)


class MultipleStorePluginValueMissing(BarbicanException):
    """Raised when a store plugin value is missing in service configuration."""
    def __init__(self, section_name):
        super(MultipleStorePluginValueMissing, self).__init__(
            u._("In section '{0}', secret_store_plugin value is missing"
                ).format(section_name)
        )
        self.section_name = section_name



# Copyright (c) 2015 Rackspace, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.
import copy

from testtools import testcase

from functionaltests.api import base
from functionaltests.api.v1.behaviors import consumer_behaviors
from functionaltests.api.v1.behaviors import container_behaviors
from functionaltests.api.v1.behaviors import secret_behaviors
from functionaltests.api.v1.models import consumer_model
from functionaltests.api.v1.models import container_models
from functionaltests.api.v1.models import secret_models

create_secret_data = {
    "name": "AES key",
    "expiration": "2030-02-28T19:14:44.180394",
    "algorithm": "aes",
    "bit_length": 256,
    "mode": "cbc",
    "payload": "gF6+lLoF3ohA9aPRpt+6bQ==",
    "payload_content_type": "application/octet-stream",
    "payload_content_encoding": "base64",
}

default_consumer_data = {
    "name": "consumername",
    "URL": "consumerURL"
}

create_generic_container_data = {
    "name": "containername",
    "type": "generic",
    "secret_refs": []
}

create_cert_container_data = {
    "name": "A Certificate Container",
    "type": "certificate",
    "secret_refs": []
}

dummy_project_id = 'dummy123'


class ConsumersBaseTestCase(base.TestCase):

    def setUp(self):
        super(ConsumersBaseTestCase, self).setUp()
        self.secret_behaviors = secret_behaviors.SecretBehaviors(self.client)
        self.container_behaviors = container_behaviors.ContainerBehaviors(
            self.client
        )
        self.consumer_behaviors = consumer_behaviors.ConsumerBehaviors(
            self.client
        )

        self.consumer_data = copy.deepcopy(default_consumer_data)

        self.generic_container_ref = self._create_populated_generic_container()

    def tearDown(self):
        self.secret_behaviors.delete_all_created_secrets()
        self.container_behaviors.delete_all_created_containers()
        self.consumer_behaviors.delete_all_created_consumers()
        super(ConsumersBaseTestCase, self).tearDown()

    def _create_a_secret(self):
        secret_model = secret_models.SecretModel(**create_secret_data)
        resp, secret_ref = self.secret_behaviors.create_secret(secret_model)
        self.assertEqual(201, resp.status_code)
        self.assertIsNotNone(secret_ref)
        return secret_ref

    def _add_secret_ref_to_container(self, container, name, ref):
        container['secret_refs'].append({"name": name, "secret_ref": ref})

    def _create_populated_generic_container(self):
        # Set up two secrets
        secret_ref_1 = self._create_a_secret()
        secret_ref_2 = self._create_a_secret()

        # Create a generic container with our secrets

        generic_container_data = copy.deepcopy(create_generic_container_data)

        self._add_secret_ref_to_container(generic_container_data,
                                          'secret_ref_1', secret_ref_1)
        self._add_secret_ref_to_container(generic_container_data,
                                          'secret_ref_2', secret_ref_2)

        container_model = container_models.ContainerModel(
            **generic_container_data
        )

        resp, container_ref = self.container_behaviors.create_container(
            container_model
        )
        self.assertEqual(201, resp.status_code)
        self.assertIsNotNone(container_ref)

        return container_ref


class ConsumersCertContainerTestCase(ConsumersBaseTestCase):

    def setUp(self):
        super(ConsumersCertContainerTestCase, self).setUp()

        self.container_default_data = copy.deepcopy(create_cert_container_data)
        self.consumer_default_data = copy.deepcopy(default_consumer_data)

    def _create_consumer(self, container_ref):
        self.consumer_test_model = consumer_model.ConsumerModel(
            **self.consumer_default_data)
        resp, consumer_data = self.consumer_behaviors.create_consumer(
            self.consumer_test_model, container_ref)

        self.assertEqual(200, resp.status_code)
        self.assertIsNotNone(consumer_data)
        return consumer_data

    def _get_decrypted_secrets_from_container(self, container_href):
        get_resp = self.container_behaviors.get_container(container_href)
        self.assertEqual(200, get_resp.status_code)

        private_key_ref = get_resp.model.secret_refs[0].secret_ref
        tls_cert_ref = get_resp.model.secret_refs[1].secret_ref
        passphrase_ref = get_resp.model.secret_refs[2].secret_ref
        intermediates_ref = get_resp.model.secret_refs[3].secret_ref

        private_key = self.secret_behaviors.get_secret(
            private_key_ref, 'application/octet-stream')
        tls_cert = self.secret_behaviors.get_secret(
            tls_cert_ref, 'application/octet-stream')
        passphrase = self.secret_behaviors.get_secret(
            passphrase_ref, 'application/octet-stream')
        intermediates = self.secret_behaviors.get_secret(
            intermediates_ref, 'application/octet-stream')

        return private_key, tls_cert, passphrase, intermediates

    def _create_populated_cert_container(self):
        dummy_private_key_ref = self._create_a_secret()
        dummy_tls_certificate_ref = self._create_a_secret()
        dummy_passphrase_ref = self._create_a_secret()
        dummy_intermediates_ref = self._create_a_secret()
        container_ref = self._create_cert_container(dummy_private_key_ref,
                                                    dummy_tls_certificate_ref,
                                                    dummy_passphrase_ref,
                                                    dummy_intermediates_ref)
        return container_ref

    def _create_cert_container(self, private_key_ref, tls_certificate_ref,
                               passphrase_ref=None, intermediates_ref=None):

        container_data = copy.deepcopy(self.container_default_data)
        self._add_secret_ref_to_container(container_data, "certificate",
                                          tls_certificate_ref)
        self._add_secret_ref_to_container(container_data, "private_key",
                                          private_key_ref)

        if passphrase_ref:
            self._add_secret_ref_to_container(container_data,
                                              "private_key_passphrase",
                                              passphrase_ref)

        if intermediates_ref:
            self._add_secret_ref_to_container(container_data, "intermediates",
                                              intermediates_ref)

        test_model = container_models.ContainerModel(
            **container_data)

        resp, container_ref = self.container_behaviors.create_container(
            test_model)
        self.assertEqual(201, resp.status_code)
        self.assertIsNotNone(container_ref)
        return container_ref

    def _deregister_consumer(self, container_ref):
        resp, consumer_data = self.consumer_behaviors.delete_consumer(
            self.consumer_test_model, container_ref
        )
        self.assertEqual(200, resp.status_code)
        self.assertIsNotNone(consumer_data)
        self.assertNotIn(self.consumer_test_model.name, consumer_data)
        self.assertNotIn(self.consumer_test_model.URL, consumer_data)
        return consumer_data

    @testcase.attr('positive')
    def test_consumer_of_cert_container_full_flow(self):
        """Simulate the typical flow for a consumer of a cert container.

        First, create a container and load it up with a key, cert and
        passphrase.

        Second, register as a consumer for that container

        Third, fetch the data from that container

        Finally, deregister as a consumer for the container
        """

        # create the populated container with a cert
        container_ref = self._create_populated_cert_container()
        self.assertIsNotNone(container_ref)

        # register as a consumer here
        consumer_data = self._create_consumer(container_ref)
        self.assertIsNotNone(consumer_data)
        self.assertEqual(1, len(consumer_data))

        # fetch the cert info from the container
        pk_response, cert_response, passphrase_response, inters_response =\
            self._get_decrypted_secrets_from_container(container_ref)
        self.assertIsNotNone(pk_response)
        self.assertIsNotNone(cert_response)
        self.assertIsNotNone(passphrase_response)
        self.assertIsNotNone(inters_response)

        # deregister as a consumer
        updated_consumer_data = self._deregister_consumer(container_ref)
        self.assertIsNotNone(updated_consumer_data)
        self.assertEqual(0, len(updated_consumer_data))


class ConsumersAuthedTestCase(ConsumersBaseTestCase):

    @testcase.attr('security')
    def test_consumer_create_authed(self):
        """Create a consumer as an authenticated user

        Should return 200
        """

        model = consumer_model.ConsumerModel(**self.consumer_data)
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, use_auth=True
        )
        self.assertEqual(200, resp.status_code)


class ConsumersUnauthedTestCase(ConsumersBaseTestCase):

    @testcase.attr('negative', 'security')
    def test_consumer_create_unauthed_no_proj_id(self):
        """Attempt to create a consumer without a token or project id

        Should return 401
        """

        model = consumer_model.ConsumerModel(**self.consumer_data)
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, use_auth=False
        )
        self.assertEqual(401, resp.status_code)

    @testcase.attr('negative', 'security')
    def test_consumer_get_list_unauthed_no_proj_id(self):
        """Attempt to get the list of consumers w/o a token or project id

        Should return 401
        """

        resp, consumers, next_ref, prev_ref = (
            self.consumer_behaviors.get_consumers(
                self.generic_container_ref, use_auth=False
            )
        )

        self.assertEqual(401, resp.status_code)

    @testcase.attr('negative', 'security')
    def test_consumer_delete_unauthed_no_proj_id(self):
        """Attempt to delete a consumer without a token or project id

        Should return 401
        """

        resp, consumer_dat = self.consumer_behaviors.delete_consumer(
            None, self.generic_container_ref, use_auth=False
        )

        self.assertEqual(401, resp.status_code)

    @testcase.attr('negative', 'security')
    def test_consumer_create_unauthed_with_proj_id(self):
        """Attempt to create a consumer with a project id, but no token

        Should return 401
        """

        model = consumer_model.ConsumerModel(**self.consumer_data)
        headers = {'X-Project-Id': dummy_project_id}
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, extra_headers=headers,
            use_auth=False
        )

        self.assertEqual(401, resp.status_code)

    @testcase.attr('negative', 'security')
    def test_consumer_get_list_unauthed_with_proj_id(self):
        """Attempt to get the list of consumers with a project id but no token

        Should return 401
        """

        headers = {'X-Project-Id': dummy_project_id}
        resp, consumers, next_ref, prev_ref = (
            self.consumer_behaviors.get_consumers(
                self.generic_container_ref, extra_headers=headers,
                use_auth=False
            )
        )

        self.assertEqual(401, resp.status_code)

    @testcase.attr('negative', 'security')
    def test_consumer_delete_unauthed_with_proj_id(self):
        """Attempt to delete a consumer with a project id, but no token

        Should return 401
        """

        headers = {'X-Project-Id': dummy_project_id}
        resp, consumer_dat = self.consumer_behaviors.delete_consumer(
            None, self.generic_container_ref, extra_headers=headers,
            use_auth=False
        )

        self.assertEqual(401, resp.status_code)


class ConsumersValidationTestCase(ConsumersBaseTestCase):

    def test_consumer_create_pass(self):
        """Create a valid consumer

        Should return 200
        """
        data = {
            "name": "consumername",
            "URL": "consumerURL"
        }
        model = consumer_model.ConsumerModel(**data)
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, use_auth=True
        )
        self.assertEqual(200, resp.status_code)

    @testcase.attr('negative')
    def test_consumer_create_fail_no_name(self):
        """Attempt to create invalid consumer (Missing name)

        Should return 400
        """
        data = {
            "URL": "consumerURL"
        }
        model = consumer_model.ConsumerModel(**data)
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, use_auth=True
        )
        self.assertEqual(400, resp.status_code)

    @testcase.attr('negative')
    def test_consumer_create_fail_no_url(self):
        """Attempt to create invalid consumer (Missing URL)

        Should return 400
        """
        data = {
            "name": "consumername"
        }
        model = consumer_model.ConsumerModel(**data)
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, use_auth=True
        )
        self.assertEqual(400, resp.status_code)

    @testcase.attr('negative')
    def test_consumer_create_fail_empty_name(self):
        """Attempt to create invalid consumer (Empty name)

        Should return 400
        """
        data = {
            "name": "",
            "URL": "consumerURL"
        }
        model = consumer_model.ConsumerModel(**data)
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, use_auth=True
        )
        self.assertEqual(400, resp.status_code)

    @testcase.attr('negative')
    def test_consumer_create_fail_empty_url(self):
        """Attempt to create invalid consumer (Empty URL)

        Should return 400
        """
        data = {
            "name": "consumername",
            "URL": ""
        }
        model = consumer_model.ConsumerModel(**data)
        resp, consumer_dat = self.consumer_behaviors.create_consumer(
            model, self.generic_container_ref, use_auth=True
        )
        self.assertEqual(400, resp.status_code)



# Copyright (c) 2014 Red Hat, Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import base64
from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import serialization
import os
from oslo_utils import uuidutils
import time

import pki
import pki.client
import pki.crypto as cryptoutil
import pki.key as key
import pki.kra
import pki.profile

from barbican.common import exception
from barbican.common import utils
from barbican import i18n as u

# we want to keep the dogtag config options separated. That way we
# do not need to import every dogtag requirement to generate the
# sample config
import barbican.plugin.dogtag_config_opts  # noqa
import barbican.plugin.interface.secret_store as sstore

# reuse the conf object to not call config.new_config() twice
CONF = barbican.plugin.dogtag_config_opts.CONF
LOG = utils.getLogger(__name__)

CERT_HEADER = "-----BEGIN CERTIFICATE-----"
CERT_FOOTER = "-----END CERTIFICATE-----"

KRA_TRANSPORT_NICK = "KRA transport cert"


def _create_nss_db_if_needed(nss_db_path, nss_password):
    """Creates NSS DB if it's not setup already

    :returns: True or False whether the database was created or not.
    """
    if not os.path.exists(nss_db_path):
        cryptoutil.NSSCryptoProvider.setup_database(
            nss_db_path, nss_password, over_write=True)
        return True
    else:
        LOG.info("The nss_db_path provided already exists, so the "
                 "database is assumed to be already set up.")
        return False


def _setup_nss_db_services(conf):
    """Sets up NSS Crypto functions

    This sets up the NSSCryptoProvider and the database it needs for it to
    store certificates. If the path specified in the configuration is already
    existent, it will assume that the database is already setup.

    This will also import the transport cert needed by the KRA if the NSS DB
    was created.
    """
    nss_db_path, nss_password = (conf.dogtag_plugin.nss_db_path,
                                 conf.dogtag_plugin.nss_password)
    if nss_db_path is None:
        LOG.warning("nss_db_path was not provided so the crypto "
                    "provider functions were not initialized.")
        return None
    if nss_password is None:
        raise ValueError(u._("nss_password is required"))
    if type(nss_password) is not bytes:
        # Password needs to be a bytes object in Python 3
        nss_password = nss_password.encode('UTF-8')

    nss_db_created = _create_nss_db_if_needed(nss_db_path, nss_password)
    crypto = cryptoutil.NSSCryptoProvider(nss_db_path, nss_password)
    if nss_db_created:
        _import_kra_transport_cert_to_nss_db(conf, crypto)

    return crypto


def _import_kra_transport_cert_to_nss_db(conf, crypto):
    try:
        connection = create_connection(conf, 'kra')
        kraclient = pki.kra.KRAClient(connection, crypto)
        systemcert_client = kraclient.system_certs

        transport_cert = systemcert_client.get_transport_cert()
        crypto.import_cert(KRA_TRANSPORT_NICK, transport_cert, ",,")
    except Exception as e:
        LOG.debug("Error importing KRA transport cert.", exc_info=True)
        LOG.error("Error in importing transport cert."
                  " KRA may not be enabled: %s", e)


def create_connection(conf, subsystem_path):
    pem_path = conf.dogtag_plugin.pem_path
    if pem_path is None:
        raise ValueError(u._("pem_path is required"))
    # port is string type in PKIConnection
    connection = pki.client.PKIConnection(
        'https',
        conf.dogtag_plugin.dogtag_host,
        str(conf.dogtag_plugin.dogtag_port),
        subsystem_path)
    connection.set_authentication_cert(pem_path)
    return connection


crypto = _setup_nss_db_services(CONF)
if crypto:
    crypto.initialize()


class DogtagPluginAlgorithmException(exception.BarbicanException):
    message = u._("Invalid algorithm passed in")


class DogtagPluginNotSupportedException(exception.NotSupported):
    message = u._("Operation not supported by Dogtag Plugin")

    def __init__(self, msg=None):
        if not msg:
            message = self.message
        else:
            message = msg

        super(DogtagPluginNotSupportedException, self).__init__(message)


class DogtagPluginArchivalException(exception.BarbicanException):
    message = u._("Key archival failed.  Error returned from KRA.")


class DogtagPluginGenerationException(exception.BarbicanException):
    message = u._("Key generation failed.  Error returned from KRA.")


class DogtagKRAPlugin(sstore.SecretStoreBase):
    """Implementation of the secret store plugin with KRA as the backend."""

    # metadata constants
    ALG = "alg"
    BIT_LENGTH = "bit_length"
    GENERATED = "generated"
    KEY_ID = "key_id"
    SECRET_MODE = "secret_mode"  # nosec
    PASSPHRASE_KEY_ID = "passphrase_key_id"  # nosec
    CONVERT_TO_PEM = "convert_to_pem"

    # string constants
    DSA_PRIVATE_KEY_HEADER = '-----BEGIN DSA PRIVATE KEY-----'
    DSA_PRIVATE_KEY_FOOTER = '-----END DSA PRIVATE KEY-----'
    DSA_PUBLIC_KEY_HEADER = '-----BEGIN DSA PUBLIC KEY-----'
    DSA_PUBLIC_KEY_FOOTER = '-----END DSA PUBLIC KEY-----'

    def __init__(self, conf=CONF):
        """Constructor - create the keyclient."""
        LOG.debug("starting DogtagKRAPlugin init")
        connection = create_connection(conf, 'kra')

        # create kraclient
        kraclient = pki.kra.KRAClient(connection, crypto)
        self.keyclient = kraclient.keys

        self.keyclient.set_transport_cert(KRA_TRANSPORT_NICK)
        self.plugin_name = conf.dogtag_plugin.plugin_name
        self.retries = conf.dogtag_plugin.retries

        LOG.debug("completed DogtagKRAPlugin init")

    def get_plugin_name(self):
        return self.plugin_name

    def store_secret(self, secret_dto):
        """Store a secret in the KRA

        If secret_dto.transport_key is not None, then we expect
        secret_dto.secret to include a base64 encoded PKIArchiveOptions
        structure as defined in section 6.4 of RFC 2511. This package contains
        a transport key wrapped session key, the session key wrapped secret
        and parameters to specify the symmetric key wrapping.

        Otherwise, the data is unencrypted and we use a call to archive_key()
        to have the Dogtag KRA client generate the relevant session keys.

        The secret_dto contains additional information on the type of secret
        that is being stored.  We will use that shortly.  For, now, lets just
        assume that its all PASS_PHRASE_TYPE

        Returns a dict with the relevant metadata (which in this case is just
        the key_id
        """
        data_type = key.KeyClient.PASS_PHRASE_TYPE
        key_id = None

        attempts = 0
        offset_time = 1
        while attempts <= self.retries and key_id is None:
            client_key_id = uuidutils.generate_uuid(dashed=False)
            if secret_dto.transport_key is not None:
                # TODO(alee-3) send the transport key with the archival request
                # once the Dogtag Client API changes.
                response = self.keyclient.archive_pki_options(
                    client_key_id,
                    data_type,
                    secret_dto.secret,
                    key_algorithm=None,
                    key_size=None)
            else:
                response = self.keyclient.archive_key(
                    client_key_id,
                    data_type,
                    secret_dto.secret,
                    key_algorithm=None,
                    key_size=None)

            key_id = response.get_key_id()

            if key_id is None:
                LOG.warning("key_id is None.  attempts: {}".format(attempts))
                attempts += 1
                time.sleep(offset_time)
                offset_time += 1

        if key_id is None:
            raise DogtagPluginArchivalException

        meta_dict = {DogtagKRAPlugin.KEY_ID: key_id}

        self._store_secret_attributes(meta_dict, secret_dto)
        return meta_dict

    def get_secret(self, secret_type, secret_metadata):
        """Retrieve a secret from the KRA

        The secret_metadata is simply the dict returned by a store_secret() or
        get_secret() call.  We will extract the key_id from this dict.

        Note: There are two ways to retrieve secrets from the KRA.

        The first method calls retrieve_key without a wrapping key.  This
        relies on the KRA client to generate a wrapping key (and wrap it with
        the KRA transport cert), and is completely transparent to the
        Barbican server.  What is returned to the caller is the
        unencrypted secret.

        The second way is to provide a wrapping key that would be generated
        on the barbican client.  That way only the client will be
        able to unwrap the secret.  This wrapping key is provided in the
        secret_metadata by Barbican core.

        Format/Type of the secret returned in the SecretDTO object.
        -----------------------------------------------------------
        The type of the secret returned is always dependent on the way it is
        stored using the store_secret method.

        In case of strings - like passphrase/PEM strings, the return will be a
        string.

        In case of binary data - the return will be the actual binary data.

        In case of retrieving an asymmetric key that is generated using the
        dogtag plugin, then the binary representation of, the asymmetric key in
        PEM format, is returned
        """
        key_id = secret_metadata[DogtagKRAPlugin.KEY_ID]

        key_spec = sstore.KeySpec(
            alg=secret_metadata.get(DogtagKRAPlugin.ALG, None),
            bit_length=secret_metadata.get(DogtagKRAPlugin.BIT_LENGTH, None),
            mode=secret_metadata.get(DogtagKRAPlugin.SECRET_MODE, None),
            passphrase=None
        )

        generated = secret_metadata.get(DogtagKRAPlugin.GENERATED, False)

        passphrase = self._get_passphrase_for_a_private_key(
            secret_type, secret_metadata, key_spec)

        recovered_key = None
        twsk = DogtagKRAPlugin._get_trans_wrapped_session_key(secret_type,
                                                              secret_metadata)

        if DogtagKRAPlugin.CONVERT_TO_PEM in secret_metadata:
            # Case for returning the asymmetric keys generated in KRA.
            # Asymmetric keys generated in KRA are not generated in PEM format.
            # This marker DogtagKRAPlugin.CONVERT_TO_PEM is set in the
            # secret_metadata for asymmetric keys generated in KRA to
            # help convert the returned private/public keys to PEM format and
            # eventually return the binary data of the keys in PEM format.

            if secret_type == sstore.SecretType.PUBLIC:
                # Public key should be retrieved using the get_key_info method
                # as it is treated as an attribute of the asymmetric key pair
                # stored in the KRA database.

                key_info = self.keyclient.get_key_info(key_id)
                recovered_key = serialization.load_der_public_key(
                    key_info.public_key,
                    backend=default_backend()
                ).public_bytes(
                    serialization.Encoding.PEM,
                    serialization.PublicFormat.PKCS1)

            elif secret_type == sstore.SecretType.PRIVATE:
                key_data = self.keyclient.retrieve_key(key_id)
                private_key = serialization.load_der_private_key(
                    key_data.data,
                    password=None,
                    backend=default_backend()
                )

                if passphrase is not None:
                    e_alg = serialization.BestAvailableEncryption(passphrase)
                else:
                    e_alg = serialization.NoEncryption()

                recovered_key = private_key.private_bytes(
                    encoding=serialization.Encoding.PEM,
                    format=serialization.PrivateFormat.PKCS8,
                    encryption_algorithm=e_alg
                )
        else:
            # TODO(alee-3) send transport key as well when dogtag client API
            # changes in case the transport key has changed.
            key_data = self.keyclient.retrieve_key(key_id, twsk)
            if twsk:
                # The data returned is a byte array.
                recovered_key = key_data.encrypted_data
            else:
                recovered_key = key_data.data

        # TODO(alee) remove final field when content_type is removed
        # from secret_dto

        if generated:
            recovered_key = base64.b64encode(recovered_key)

        ret = sstore.SecretDTO(
            type=secret_type,
            secret=recovered_key,
            key_spec=key_spec,
            content_type=None,
            transport_key=None)

        return ret

    def delete_secret(self, secret_metadata):
        """Delete a secret from the KRA

        There is currently no way to delete a secret in Dogtag.
        We will be implementing such a method shortly.
        """
        pass

    def generate_symmetric_key(self, key_spec):
        """Generate a symmetric key

        This calls generate_symmetric_key() on the KRA passing in the
        algorithm, bit_length and id (used as the client_key_id) from
        the secret.  The remaining parameters are not used.

        Returns a metadata object that can be used for retrieving the secret.
        """

        usages = [key.SymKeyGenerationRequest.DECRYPT_USAGE,
                  key.SymKeyGenerationRequest.ENCRYPT_USAGE]

        algorithm = self._map_algorithm(key_spec.alg.lower())

        if algorithm is None:
            raise DogtagPluginAlgorithmException
        passphrase = key_spec.passphrase
        if passphrase:
            raise DogtagPluginNotSupportedException(
                u._("Passphrase encryption is not supported for symmetric"
                    " key generating algorithms."))

        key_id = None
        attempts = 0
        offset_time = 1
        while attempts <= self.retries and key_id is None:
            client_key_id = uuidutils.generate_uuid()
            response = self.keyclient.generate_symmetric_key(
                client_key_id,
                algorithm,
                key_spec.bit_length,
                usages)
            key_id = response.get_key_id()

            if key_id is None:
                LOG.warning("generate_symkey: key_id is None.  attempts: {}"
                            .format(attempts))
                attempts += 1
                time.sleep(offset_time)
                offset_time += 1

        if key_id is None:
            raise DogtagPluginGenerationException

        # Barbican expects stored keys to be base 64 encoded.  We need to
        # add flag to the keyclient.generate_symmetric_key() call above
        # to ensure that the key that is stored is base64 encoded.
        #
        # As a workaround until that update is available, we will store a
        # parameter "generated"  to indicate that the response must be base64
        # encoded on retrieval.  Note that this will not work for transport
        # key encoded data.
        return {DogtagKRAPlugin.ALG: key_spec.alg,
                DogtagKRAPlugin.BIT_LENGTH: key_spec.bit_length,
                DogtagKRAPlugin.KEY_ID: response.get_key_id(),
                DogtagKRAPlugin.GENERATED: True}

    def generate_asymmetric_key(self, key_spec):
        """Generate an asymmetric key.

        Note that barbican expects all secrets to be base64 encoded.
        """

        usages = [key.AsymKeyGenerationRequest.DECRYPT_USAGE,
                  key.AsymKeyGenerationRequest.ENCRYPT_USAGE]

        client_key_id = uuidutils.generate_uuid()
        algorithm = self._map_algorithm(key_spec.alg.lower())
        passphrase = key_spec.passphrase

        if algorithm is None:
            raise DogtagPluginAlgorithmException

        passphrase_key_id = None
        passphrase_metadata = None
        if passphrase:
            if algorithm == key.KeyClient.DSA_ALGORITHM:
                raise DogtagPluginNotSupportedException(
                    u._("Passphrase encryption is not "
                        "supported for DSA algorithm")
                )

            stored_passphrase_info = self.keyclient.archive_key(
                uuidutils.generate_uuid(),
                self.keyclient.PASS_PHRASE_TYPE,
                base64.b64encode(passphrase))

            passphrase_key_id = stored_passphrase_info.get_key_id()
            passphrase_metadata = {
                DogtagKRAPlugin.KEY_ID: passphrase_key_id
            }

        # Barbican expects stored keys to be base 64 encoded.  We need to
        # add flag to the keyclient.generate_asymmetric_key() call above
        # to ensure that the key that is stored is base64 encoded.
        #
        # As a workaround until that update is available, we will store a
        # parameter "generated"  to indicate that the response must be base64
        # encoded on retrieval.  Note that this will not work for transport
        # key encoded data.

        response = self.keyclient.generate_asymmetric_key(
            client_key_id,
            algorithm,
            key_spec.bit_length,
            usages)

        public_key_metadata = {
            DogtagKRAPlugin.ALG: key_spec.alg,
            DogtagKRAPlugin.BIT_LENGTH: key_spec.bit_length,
            DogtagKRAPlugin.KEY_ID: response.get_key_id(),
            DogtagKRAPlugin.CONVERT_TO_PEM: "true",
            DogtagKRAPlugin.GENERATED: True
        }

        private_key_metadata = {
            DogtagKRAPlugin.ALG: key_spec.alg,
            DogtagKRAPlugin.BIT_LENGTH: key_spec.bit_length,
            DogtagKRAPlugin.KEY_ID: response.get_key_id(),
            DogtagKRAPlugin.CONVERT_TO_PEM: "true",
            DogtagKRAPlugin.GENERATED: True
        }

        if passphrase_key_id:
            private_key_metadata[DogtagKRAPlugin.PASSPHRASE_KEY_ID] = (
                passphrase_key_id
            )

        return sstore.AsymmetricKeyMetadataDTO(private_key_metadata,
                                               public_key_metadata,
                                               passphrase_metadata)

    def generate_supports(self, key_spec):
        """Key generation supported?

        Specifies whether the plugin supports key generation with the
        given key_spec.

        For now, we will just check the algorithm.  When dogtag adds a
        call to check the bit length as well, we will use that call to
        take advantage of the bit_length information
        """
        return self._map_algorithm(key_spec.alg) is not None

    def store_secret_supports(self, key_spec):
        """Key storage supported?

        Specifies whether the plugin supports storage of the secret given
        the attributes included in the KeySpec
        """
        return True

    @staticmethod
    def _map_algorithm(algorithm):
        """Map Barbican algorithms to Dogtag plugin algorithms.

        Note that only algorithms supported by Dogtag will be mapped.
        """
        if algorithm is None:
            return None

        if algorithm.lower() == sstore.KeyAlgorithm.AES.lower():
            return key.KeyClient.AES_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DES.lower():
            return key.KeyClient.DES_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DESEDE.lower():
            return key.KeyClient.DES3_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DSA.lower():
            return key.KeyClient.DSA_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.RSA.lower():
            return key.KeyClient.RSA_ALGORITHM
        elif algorithm.lower() == sstore.KeyAlgorithm.DIFFIE_HELLMAN.lower():
            # may be supported, needs to be tested
            return None
        elif algorithm.lower() == sstore.KeyAlgorithm.EC.lower():
            # asymmetric keys not yet supported
            return None
        else:
            return None

    @staticmethod
    def _store_secret_attributes(meta_dict, secret_dto):
        # store the following attributes for retrieval
        key_spec = secret_dto.key_spec
        if key_spec is not None:
            if key_spec.alg is not None:
                meta_dict[DogtagKRAPlugin.ALG] = key_spec.alg
            if key_spec.bit_length is not None:
                meta_dict[DogtagKRAPlugin.BIT_LENGTH] = key_spec.bit_length
            if key_spec.mode is not None:
                meta_dict[DogtagKRAPlugin.SECRET_MODE] = key_spec.mode

    def _get_passphrase_for_a_private_key(self, secret_type, secret_metadata,
                                          key_spec):
        """Retrieve the passphrase for the private key stored in the KRA."""
        if secret_type is None:
            return None
        if key_spec.alg is None:
            return None

        passphrase = None
        if DogtagKRAPlugin.PASSPHRASE_KEY_ID in secret_metadata:
            if key_spec.alg.upper() == key.KeyClient.RSA_ALGORITHM:
                passphrase = self.keyclient.retrieve_key(
                    secret_metadata.get(DogtagKRAPlugin.PASSPHRASE_KEY_ID)
                ).data
            else:
                if key_spec.alg.upper() == key.KeyClient.DSA_ALGORITHM:
                    raise sstore.SecretGeneralException(
                        u._("DSA keys should not have a passphrase in the"
                            " database, for being used during retrieval.")
                    )
                raise sstore.SecretGeneralException(
                    u._("Secrets of type {secret_type} should not have a "
                        "passphrase in the database, for being used during "
                        "retrieval.").format(secret_type=secret_type)
                )

        # note that Barbican expects the passphrase to be base64 encoded when
        # stored, so we need to decode it.
        if passphrase:
            passphrase = base64.b64decode(passphrase)
        return passphrase

    @staticmethod
    def _get_trans_wrapped_session_key(secret_type, secret_metadata):
        twsk = secret_metadata.get('trans_wrapped_session_key', None)
        if secret_type in [sstore.SecretType.PUBLIC,
                           sstore.SecretType.PRIVATE]:
            if twsk:
                raise DogtagPluginNotSupportedException(
                    u._("Encryption using session key is not supported when "
                        "retrieving a {secret_type} "
                        "key.").format(secret_type=secret_type)
                )

        return twsk
